#
#  Copyright (c) 2020 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from unittest import mock
from unittest import TestCase

from netcontrold.lib import dataif
import copy


# A noop handler for netcontrold logging.
class NlogNoop(object):

    def info(self, *args):
        prefix = "%s> " % (self.__class__.__name__)
        print("%s %s" % (prefix, "".join(args)))

    def debug(self, *args):
        prefix = "%s> " % (self.__class__.__name__)
        print("%s %s" % (prefix, "".join(args)))


def mock_pmd_stats(*args):
    stats = """pmd thread numa_id 0 core_id 1:
  packets received: 1000
  packet recirculations: 0
  avg. datapath passes per packet: 1.00
  emc hits: 12768883657
  smc hits: 0
  megaflow hits: 49909
  avg. subtable lookups per megaflow hit: 1.28
  miss with success upcall: 3911
  miss with failed upcall: 0
  avg. packets per output batch: 9.37
  idle cycles: 1100 (93.95%)
  processing cycles: 1200 (6.05%)
  avg cycles per packet: 13414.81 (171292880970534/12768937477)
  avg processing cycles per packet: 812.16 (10370482753684/12768937477)
pmd thread numa_id 0 core_id 13:
  packets received: 3000
  packet recirculations: 0
  avg. datapath passes per packet: 1.00
  emc hits: 31402809758
  smc hits: 0
  megaflow hits: 14434
  avg. subtable lookups per megaflow hit: 1.48
  miss with success upcall: 3637
  miss with failed upcall: 0
  avg. packets per output batch: 27.05
  idle cycles: 3100 (87.83%)
  processing cycles: 3200 (12.17%)
  avg cycles per packet: 5454.69 (171292606842168/31402827829)
  avg processing cycles per packet: 663.88 (20847810403960/31402827829)
main thread:
  packets received: 108
  packet recirculations: 0
  avg. datapath passes per packet: 1.00
  emc hits: 0
  smc hits: 0
  megaflow hits: 34
  avg. subtable lookups per megaflow hit: 1.00
  miss with success upcall: 74
  miss with failed upcall: 0
  avg. packets per output batch: 1.00"""  # noqa: E501
    return stats


def mock_pmd_rxqs(*args):
    stats = """pmd thread numa_id 0 core_id 1:
  isolated : false
  port: port1   queue-id:  0  pmd usage:  0 %
pmd thread numa_id 0 core_id 13:
  isolated : false
  port: port2   queue-id:  0  pmd usage:  0 %"""  # noqa: E501
    return stats


def mock_port_stats(*args):
    stats = """netdev@ovs-netdev:
  lookups: hit:0 missed:0 lost:0
  flows: 0
  port 1: port1 (tap)
    RX packets:5 errors:0 dropped:2 overruns:0 frame:0
    TX packets:5 errors:0 dropped:3 aborted:0 carrier:0
    collisions:0
    RX bytes:0  TX bytes:0
  port 2: port2 (tap)
    RX packets:5 errors:0 dropped:2 overruns:0 frame:0
    TX packets:5 errors:0 dropped:3 aborted:0 carrier:0
    collisions:0
    RX bytes:0  TX bytes:0"""  # noqa: E501
    return stats


def mock_interface_stats(*args):
    stats = """_uuid               : 583d9020-a49a-4c5d-902d-dfcaa41e2911
admin_state         : up
bfd                 : {}
bfd_status          : {}
cfm_fault           : []
cfm_fault_status    : []
cfm_flap_count      : []
cfm_health          : []
cfm_mpid            : []
cfm_remote_mpids    : []
cfm_remote_opstate  : []
duplex              : []
error               : []
external_ids        : {}
ifindex             : 6656773
ingress_policing_burst: 0
ingress_policing_rate: 0
lacp_current        : []
link_resets         : 0
link_speed          : []
link_state          : down
lldp                : {}
mac                 : []
mac_in_use          : "00:00:00:00:00:00"
mtu                 : 1500
mtu_request         : []
name                : "port1"
ofport              : 201
ofport_request      : 201
options             : {n_rxq="1", n_txq="1", vhost-server-path="/var/lib/vhost_sockets/dpdkvhost2"}
other_config        : {}
statistics          : {"rx_1024_to_1522_packets"=0, "rx_128_to_255_packets"=0, "rx_1523_to_max_packets"=0, "rx_1_to_64_packets"=0, "rx_256_to_511_packets"=0, "rx_512_to_1023_packets"=0, "rx_65_to_127_packets"=0, rx_bytes=0, rx_dropped=0, rx_errors=0, rx_packets=0, tx_bytes=0, tx_dropped=0, tx_packets=0, tx_retries=0}
status              : {mode=client, status=disconnected}
type                : dpdkvhostuserclient

_uuid               : 3f97b403-2fe3-490f-8b24-8f9b80eb7aed
admin_state         : up
bfd                 : {}
bfd_status          : {}
cfm_fault           : []
cfm_fault_status    : []
cfm_flap_count      : []
cfm_health          : []
cfm_mpid            : []
cfm_remote_mpids    : []
cfm_remote_opstate  : []
duplex              : full
error               : []
external_ids        : {}
ifindex             : 6054325
ingress_policing_burst: 0
ingress_policing_rate: 0
lacp_current        : []
link_resets         : 0
link_speed          : 10000000000
link_state          : up
lldp                : {}
mac                 : []
mac_in_use          : "24:6e:96:c4:2f:ea"
mtu                 : 1500
mtu_request         : []
name                : "port2"
ofport              : 11
ofport_request      : 11
options             : {dpdk-devargs="0000:19:00.1", n_rxq="1", n_rxq_desc="4096", n_txq="1", n_txq_desc="4096"}
other_config        : {}
statistics          : {flow_director_filter_add_errors=0, flow_director_filter_remove_errors=0, mac_local_errors=1, mac_remote_errors=0, "rx_128_to_255_packets"=0, "rx_1_to_64_packets"=0, "rx_256_to_511_packets"=0, "rx_512_to_1023_packets"=0, "rx_65_to_127_packets"=0, rx_broadcast_packets=0, rx_bytes=0, rx_crc_errors=0, rx_dropped=0, rx_errors=0, rx_fcoe_crc_errors=0, rx_fcoe_dropped=0, rx_fcoe_mbuf_allocation_errors=0, rx_fragment_errors=0, rx_illegal_byte_errors=0, rx_jabber_errors=0, rx_length_errors=0, rx_mac_short_packet_dropped=0, rx_management_dropped=0, rx_management_packets=0, rx_mbuf_allocation_errors=0, rx_missed_errors=0, rx_oversize_errors=0, rx_packets=0, "rx_priority0_dropped"=0, "rx_priority0_mbuf_allocation_errors"=0, "rx_priority1_dropped"=0, "rx_priority1_mbuf_allocation_errors"=0, "rx_priority2_dropped"=0, "rx_priority2_mbuf_allocation_errors"=0, "rx_priority3_dropped"=0, "rx_priority3_mbuf_allocation_errors"=0, "rx_priority4_dropped"=0, "rx_priority4_mbuf_allocation_errors"=0, "rx_priority5_dropped"=0, "rx_priority5_mbuf_allocation_errors"=0, "rx_priority6_dropped"=0, "rx_priority6_mbuf_allocation_errors"=0, "rx_priority7_dropped"=0, "rx_priority7_mbuf_allocation_errors"=0, rx_undersize_errors=0, "tx_128_to_255_packets"=0, "tx_1_to_64_packets"=0, "tx_256_to_511_packets"=0, "tx_512_to_1023_packets"=0, "tx_65_to_127_packets"=0, tx_broadcast_packets=0, tx_bytes=0, tx_dropped=0, tx_errors=0, tx_management_packets=0, tx_multicast_packets=0, tx_packets=0}
status              : {driver_name=net_ixgbe, if_descr="DPDK 18.11.2 net_ixgbe", if_type="6", link_speed="10Gbps", max_hash_mac_addrs="4096", max_mac_addrs="127", max_rx_pktlen="1518", max_rx_queues="128", max_tx_queues="64", max_vfs="0", max_vmdq_pools="64", min_rx_bufsize="1024", numa_id="0", pci-device_id="0x10fb", pci-vendor_id="0x8086", port_no="1"}
type                : dpdk"""  # noqa: E501
    return stats


class TestDataif_Collection(TestCase):
    """
    Test for getting pmd stats.
    """

    # create an empty pmd_map
    pmd_map = dict()

    def setUp(self):
        # turn off limited info shown in assert failure for pmd object.
        self.maxDiff = None

        dataif.Context.nlog = NlogNoop()

        # create one pmd object.
        fx_pmd_1 = dataif.Dataif_Pmd(1)

        # let it be in numa 0.
        fx_pmd_1.numa_id = 0

        # add it to pmd_map
        self.pmd_map[1] = fx_pmd_1

        # create port class of name 'port1'.
        dataif.make_dataif_port("port1")

        # add port object into pmd.
        port1 = fx_pmd_1.add_port("port1")
        port1.numa_id = fx_pmd_1.numa_id

        # add rxq to port object
        rxq_1 = port1.add_rxq(0)  # noqa: F841

        # create another pmd object.
        fx_pmd_2 = dataif.Dataif_Pmd(13)

        # let it be in numa 0.
        fx_pmd_2.numa_id = 0

        # add it to pmd_map
        self.pmd_map[13] = fx_pmd_2

        # create port class of name 'port2'.
        dataif.make_dataif_port("port2")

        # add port object into pmd.
        port2 = fx_pmd_2.add_port("port2")
        port2.numa_id = fx_pmd_2.numa_id

        # add rxq to port object
        rxq_2 = port2.add_rxq(0)  # noqa: F841

        return

    # Test case:
    #   getting pmd stats from get_pmd_stats function and checking if
    #   declared pmd_map is modified or not
    @mock.patch('netcontrold.lib.util.exec_host_command', mock_pmd_stats)
    def test_get_pmd_stats_1(self):
        # create a copy of original pmd_map
        expected = copy.deepcopy(self.pmd_map)

        # modify pmd_1 object to expected
        expected_pmd_1 = expected[1]
        expected_pmd_1.rx_cyc[expected_pmd_1.cyc_idx] = 1000
        expected_pmd_1.idle_cpu_cyc[expected_pmd_1.cyc_idx] = 1100
        expected_pmd_1.proc_cpu_cyc[expected_pmd_1.cyc_idx] = 1200

        # modify pmd_2 object to expected
        expected_pmd_2 = expected[13]
        expected_pmd_2.rx_cyc[expected_pmd_2.cyc_idx] = 3000
        expected_pmd_2.idle_cpu_cyc[expected_pmd_2.cyc_idx] = 3100
        expected_pmd_2.proc_cpu_cyc[expected_pmd_2.cyc_idx] = 3200

        # calling get_pmd_stats function. pmd_map is modified here
        out = dataif.get_pmd_stats(self.pmd_map)

        # modified pmd objects
        out_pmd_1 = out[1]
        out_pmd_2 = out[13]

        # check if original and modified pmd_map objects are different.
        # if __eq__ returns false , then pmd objects are modified
        self.assertEqual(
            out_pmd_1,
            expected_pmd_1,
            "pmd 1 stats to be matched")

        self.assertEqual(
            out_pmd_2,
            expected_pmd_2,
            "pmd 2 stats to be matched")

    # Test case:
    #   getting interface stats from get_interface_stats function and checking
    #   if declared port objects are modified or not
    @mock.patch('netcontrold.lib.util.exec_host_command', mock_interface_stats)
    def test_get_interface_stats_1(self):
        # create a copy of original port1
        pmd1 = self.pmd_map[1]
        port1 = pmd1.find_port_by_name('port1')
        expected_port1 = copy.deepcopy(port1)

        # modify these to expected
        expected_port1.type = "dpdkvhostuserclient"
        expected_port1.tx_retry_cyc[expected_port1.cyc_idx] = 0

        # calling get_interface_stats function.port1 object is modified here
        dataif.get_interface_stats()

        # check if expected and modified port1 objects are same
        self.assertEqual(port1, expected_port1, "port 1 to be matched")

        # create a copy of original port2
        pmd2 = self.pmd_map[13]
        port2 = pmd2.find_port_by_name('port2')
        expected_port2 = copy.deepcopy(port2)

        # modify these to expected
        expected_port2.type = "dpdk"
        expected_port2.tx_retry_cyc[expected_port2.cyc_idx] = 0

        # calling get_interface_stats function.port2 object is modified here
        dataif.get_interface_stats()

        # check if expected and modified port2 objects are same
        self.assertEqual(port2, expected_port2, "port 2 to be matched")



#!/usr/bin/env python3
#
#  Copyright (c) 2019 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

__all__ = ['ncd_main']

# import system libraries
import re
import signal
import time
import argparse
import sys
import socket
import os
import logging
import threading
import json
from logging.handlers import RotatingFileHandler
from datetime import datetime

import netcontrold
from netcontrold.lib import config
from netcontrold.lib import dataif
from netcontrold.lib import util
from netcontrold.lib import error


class RebalContext(dataif.Context):
    rebal_mode = False
    rebal_quick = False
    rebal_tick = 0
    rebal_tick_n = 0
    apply_rebal = False


class TraceContext(dataif.Context):
    trace_mode = False


nlog = None


class CtlDThread(util.Thread):

    def __init__(self, eobj):
        util.Thread.__init__(self, eobj)

    def run(self):
        sock_file = config.ncd_socket

        try:
            os.unlink(sock_file)
        except OSError:
            if os.path.exists(sock_file):
                raise

        os.makedirs(os.path.dirname(sock_file), exist_ok=True)

        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        nlog.info("starting ctld on %s" % sock_file)
        sock.bind(sock_file)
        sock.listen(1)

        while (not self.ncd_shutdown.is_set()):
            conn, client = sock.accept()

            try:
                cmd = conn.recv(24).decode()

                ctx = dataif.Context
                rctx = RebalContext
                tctx = TraceContext

                if cmd == 'CTLD_TRACE_ON':
                    if not tctx.trace_mode:
                        nlog.info("turning on trace mode ..")
                        tctx.trace_mode = True
                    else:
                        nlog.info("trace mode already on ..!")

                    conn.sendall(b"CTLD_ACK")

                elif cmd == 'CTLD_TRACE_OFF':
                    if tctx.trace_mode:
                        nlog.info("turning off trace mode ..")
                        tctx.trace_mode = False
                    else:
                        nlog.info("trace mode already off ..!")

                    conn.sendall(b"CTLD_ACK")

                elif cmd == 'CTLD_REBAL_ON':
                    if not rctx.rebal_mode:
                        nlog.info("turning on rebalance mode ..")
                        rctx.rebal_mode = True
                    else:
                        nlog.info("rebalance mode already on ..!")

                    conn.sendall(b"CTLD_ACK")

                elif cmd == 'CTLD_REBAL_OFF':
                    if rctx.rebal_mode:
                        nlog.info("turning off rebalance mode ..")
                        rctx.rebal_mode = False
                    else:
                        nlog.info("rebalance mode already off ..!")

                    conn.sendall(b"CTLD_ACK")

                elif cmd == 'CTLD_REBAL_QUICK_ON':
                    if not rctx.rebal_quick:
                        nlog.info("turning on rebalance quick mode ..")
                        rctx.rebal_quick = True
                    else:
                        nlog.info("rebalance quick mode already on ..!")

                    conn.sendall(b"CTLD_ACK")

                elif cmd == 'CTLD_REBAL_QUICK_OFF':
                    if rctx.rebal_quick:
                        nlog.info("turning off rebalance quick mode ..")
                        rctx.rebal_quick = False
                    else:
                        nlog.info("rebalance quick mode already off ..!")

                    conn.sendall(b"CTLD_ACK")

                elif cmd == 'CTLD_VERBOSE_ON':
                    fh = ctx.log_handler
                    if fh.level == logging.INFO:
                        nlog.info("turning on verbose mode ..")
                        fh.setLevel(logging.DEBUG)
                    else:
                        nlog.info("verbose mode already on ..!")

                    conn.sendall(b"CTLD_ACK")

                elif cmd == 'CTLD_VERBOSE_OFF':
                    fh = ctx.log_handler
                    if fh.level == logging.DEBUG:
                        nlog.info("turning off verbose mode ..")
                        fh.setLevel(logging.INFO)
                    else:
                        nlog.info("verbose mode already off ..!")

                    conn.sendall(b"CTLD_ACK")

                elif cmd == 'CTLD_REBAL_CNT':
                    n = 0
                    if rctx.rebal_mode:
                        n = len(rctx.rebal_stat)
                        for (x, y, z) in ctx.events:
                            if y == 'rebalance':
                                n += 1

                    conn.sendall(b"CTLD_DATA_ACK %6d" % (len(str(n))))
                    conn.sendall(str(n).encode())

                elif cmd == 'CTLD_CONFIG':
                    status = "trace mode:"
                    if tctx.trace_mode:
                        status += " on\n"
                    else:
                        status += " off\n"

                    status += "rebalance mode:"
                    if rctx.rebal_mode:
                        status += " on\n"
                    else:
                        status += " off\n"

                    status += "rebalance quick:"
                    if rctx.rebal_quick:
                        status += " on\n"
                    else:
                        status += " off\n"

                    status += "verbose log:"
                    fh = ctx.log_handler
                    if fh.level == logging.DEBUG:
                        status += " on\n"
                    else:
                        status += " off\n"

                    conn.sendall(b"CTLD_DATA_ACK %6d" % (len(status)))
                    conn.sendall(status.encode())

                elif cmd == 'CTLD_STATUS':
                    status = "%-16s | %-12s | %s\n" % ('Interface',
                                                       'Event', 'Time stamp')
                    status += ('-' * 17) + '+' + ('-' * 14) + '+' + ('-' * 28)
                    status += '\n'

                    for (x, y, z) in ctx.events:
                        status += "%-16s | %-12s | %s\n" % (x, y, z)

                    conn.sendall(b"CTLD_DATA_ACK %6d" % (len(status)))
                    conn.sendall(status.encode())

                elif cmd == 'CTLD_STATUS_CLEAR':
                    ctx.events = []
                    conn.sendall(b"CTLD_ACK")

                elif cmd == 'CTLD_VERSION':
                    status = "netcontrold v%s\n" % netcontrold.__version__
                    ret = util.exec_host_command("ovs-vsctl -V")
                    if ret == 1:
                        status += "openvswitch (unknown)\n"
                    else:
                        parse = re.match("ovs-vsctl \(Open vSwitch\) (.*?)\n",
                                         ret)
                        status += "openvswitch v%s\n" % parse[1]

                    conn.sendall(b"CTLD_DATA_ACK %6d" % (len(status)))
                    conn.sendall(status.encode())

                else:
                    nlog.info("unknown control command %s" % cmd)

            finally:
                conn.close()

        return


def collect_data(n_samples, s_sampling):
    """
    Collect various stats and rxqs mapping of every pmd in the vswitch.

    Parameters
    ----------
    n_samples : int
        number of samples

    s_sampling: int
        sampling interval
    """

    ctx = dataif.Context
    rctx = RebalContext

    # collect samples of pmd and rxq stats.
    idx_max = n_samples
    idx_gen = (o for o in range(0, idx_max))
    while True:
        try:
            next(idx_gen)
        except StopIteration:
            break

        try:
            rctx.rebal_tick += 1
            dataif.get_port_stats()
            dataif.get_interface_stats()
            dataif.get_pmd_stats(ctx.pmd_map)
            dataif.get_pmd_rxqs(ctx.pmd_map)
        except (error.OsCommandExc,
                error.ObjCreateExc,
                error.ObjConsistencyExc,
                error.ObjParseExc,
                error.ObjModelExc) as e:
            # report error event
            now = datetime.now()
            now_ts = now.strftime("%Y-%m-%d %H:%M:%S")
            if isinstance(e, error.OsCommandExc):
                nlog.warn("unable to collect data: %s" % e)
                ctx.events.append(("switch", "error", now_ts))
                raise error.NcdShutdownExc

            elif isinstance(e, error.ObjCreateExc):
                nlog.warn("unable to create object: %s" % e)
                ctx.events.append(("ncd", "error", now_ts))
                raise error.NcdShutdownExc

            elif isinstance(e, error.ObjConsistencyExc):
                nlog.warn("inconsistency in collected data: %s" % e)
                ctx.events.append(("ncd", "error", now_ts))
                raise error.NcdShutdownExc

            elif isinstance(e, error.ObjModelExc):
                nlog.warn("switch states changed: %s" % e)
                ctx.events.append(("ncd", "retry_model", now_ts))

                # reset collected data
                ctx.pmd_map.clear()
                ctx.port_to_cls.clear()
                ctx.port_to_id.clear()

                # restart iterations
                idx_gen.close()
                idx_max = config.ncd_samples_max
                idx_gen = (o for o in range(0, idx_max))
                continue

            elif isinstance(e, error.ObjParseExc):
                nlog.warn("unable to parse info: %s" % e)
                ctx.events.append(("ncd", "retry_parse", now_ts))

                # reset collected data
                ctx.pmd_map.clear()
                ctx.port_to_cls.clear()
                ctx.port_to_id.clear()

                # restart iterations
                idx_gen.close()
                idx_max = config.ncd_samples_max
                idx_gen = (o for o in range(0, idx_max))
                time.sleep(s_sampling)
                continue

            else:
                nlog.error("unhandled exception: %s" % e)
                ctx.events.append(("ncd", "exception", now_ts))
                raise error.NcdShutdownExc

        time.sleep(s_sampling)

    now = datetime.now()
    ctx.last_ts = now.strftime("%Y-%m-%d %H:%M:%S")
    dataif.update_pmd_load(ctx.pmd_map)

    return ctx


def rebalance_switch(pmd_map):
    """
    Issue appropriate actions in vswitch to rebalance.

    Parameters
    ----------
    pmd_map : dict
        mapping of pmd id and its Dataif_Pmd object.
    """

    port_to_pmdq = {}
    non_isol_pmds = []
    numa = 0
    for pmd_id, pmd in pmd_map.items():
        # leave one pmd in every numa as non-isolated.
        if pmd.numa_id == numa:
            non_isol_pmds.append(pmd)
            numa += 1
            continue

        for port_name, port in pmd.port_map.items():
            if port_name not in port_to_pmdq and len(port.rxq_map) != 0:
                port_to_pmdq[port_name] = ""
            for rxq_id in port.rxq_map:
                port_to_pmdq[port_name] += "%d:%d," % (rxq_id, pmd_id)

    # refresh ports info and check for any port removed now.
    ctx = dataif.Context
    ctx.port_to_id.clear()
    dataif.get_port_stats()
    cmd = ""
    for port_name, pmdq in port_to_pmdq.items():
        if port_name not in ctx.port_to_id:
            now = datetime.now()
            now_ts = now.strftime("%Y-%m-%d %H:%M:%S")
            nlog.info("not setting affinity for an unavailable port %s"
                      % (port_name))
            ctx.events.append((port_name, "skip", now_ts))
            continue
        cmd += "-- set Interface %s other_config:pmd-rxq-affinity=%s " % (
            port_name, pmdq)

    # ensure non-isolated pmd carry new rxqs, arriving from other pmds.
    for pmd in non_isol_pmds:
        for port_name, port in pmd.port_map.items():
            if port_name not in ctx.port_to_id:
                now = datetime.now()
                now_ts = now.strftime("%Y-%m-%d %H:%M:%S")
                nlog.info("not resetting affinity for unavailable port %s"
                          % (port_name))
                ctx.events.append((port_name, "skip", now_ts))
                continue
            if port_name not in port_to_pmdq:
                cmd += "-- remove Interface %s other_config pmd-rxq-affinity "\
                    % (port_name)
    return "ovs-vsctl --no-wait %s" % cmd


def ncd_kill(signal, frame):
    ctx = dataif.Context
    rctx = RebalContext
    tctx = TraceContext

    nlog.critical("Got signal %s, doing required clean up .." % signal)

    # reset rebalance settings in ports
    cmd = ""
    for port_name, port in ctx.port_to_cls.items():
        # skip port that we did not rebalance.
        if not port.rebalance:
            continue

        cmd += "-- remove Interface %s other_config pmd-rxq-affinity " % (
            port_name)

    if cmd:
        ret = util.exec_host_command("ovs-vsctl --no-wait %s" % cmd)
        if ret == 0:
            nlog.info("removed pmd-rxq-affinity in rebalanced ports.")
        else:
            nlog.warn("removing pmd-rxq-affinity failed for some ports.")
            nlog.warn("you may check ovs-vsctl --no-wait %s" % cmd)

    os.makedirs(os.path.dirname(config.ncd_dump_file), exist_ok=True)
    with open(config.ncd_dump_file, 'w') as f:
        fh = ctx.log_handler
        json.dump({
            "time": "%s" % datetime.now(),
            "version": netcontrold.__version__,
            "events": ctx.events,
            "config": {
                "trace_mode": tctx.trace_mode,
                "rebalance_mode": rctx.rebal_mode,
                "verbose_log": fh.level == logging.DEBUG,
            },
        }, f)

        nlog.info("saved current configuration and data in %s" %
                  config.ncd_dump_file)

    raise error.NcdShutdownExc


def ncd_main(argv):
    # input options
    argpobj = argparse.ArgumentParser(
        prog='ncd.py', description='control network load on pmd')

    argpobj.add_argument('-s', '--sample-interval',
                         type=int,
                         default=10,
                         help='seconds between each sampling (default: 10)')

    argpobj.add_argument('-t', '--trace',
                         required=False,
                         action='store_true',
                         default=False,
                         help='operate in trace mode',
                         )

    argpobj.add_argument('--trace-cb',
                         type=str,
                         default='ncd_cb_pktdrop',
                         help='trace mode callback '
                         '(default: ncd_cb_pktdrop)')

    argpobj.add_argument('-r', '--rebalance',
                         required=False,
                         action='store_true',
                         default=True,
                         help="operate in rebalance mode",
                         )

    argpobj.add_argument('--rebalance-interval',
                         type=int,
                         default=60,
                         help='seconds between each re-balance '
                         '(default: 60)')

    argpobj.add_argument('--rebalance-n',
                         type=int,
                         default=1,
                         help='rebalance dry-runs at the max (default: 1)')

    argpobj.add_argument('--rebalance-iq',
                         action='store_true',
                         default=False,
                         help='rebalance by iterative queues logic '
                                '(default: False)')

    argpobj.add_argument('-q', '--quiet',
                         action='store_true',
                         default=False,
                         help='no logging in terminal (default: False)')

    argpobj.add_argument('-v', '--verbose',
                         action='store_true',
                         default=False,
                         help='trace logging (default: False)')

    args = argpobj.parse_args(argv)

    # check input to ncd
    ncd_trace = args.trace
    ncd_trace_cb = args.trace_cb
    ncd_rebal = args.rebalance

    if ncd_trace and not util.exists(ncd_trace_cb):
        print("no such program %s exists!" % ncd_trace_cb)
        sys.exit(1)

    # set verbose level
    os.makedirs(os.path.dirname(config.ncd_log_file), exist_ok=True)
    fh = RotatingFileHandler(config.ncd_log_file,
                             maxBytes=(config.ncd_log_max_KB * 1024),
                             backupCount=config.ncd_log_max_backup_n)

    fh_fmt = logging.Formatter(
        "%(asctime)s|%(name)s|%(levelname)s|%(message)s")
    fh.setFormatter(fh_fmt)
    if args.verbose:
        fh.setLevel(logging.DEBUG)
    else:
        fh.setLevel(logging.INFO)

    ch = logging.StreamHandler(sys.stdout)
    ch_fmt = logging.Formatter("%(message)s")
    ch.setFormatter(ch_fmt)
    ch.setLevel(logging.INFO)

    global nlog
    nlog = logging.getLogger('ncd')
    nlog.setLevel(logging.DEBUG)
    nlog.addHandler(fh)
    if not args.quiet:
        nlog.addHandler(ch)

    ctx = dataif.Context
    ctx.nlog = nlog
    ctx.log_handler = fh
    pmd_map = ctx.pmd_map

    # set sampling interval to collect data
    ncd_sample_interval = args.sample_interval

    # set interval between each re-balance
    ncd_rebal_interval = args.rebalance_interval

    # set rebalance dryrun count
    ncd_rebal_n = args.rebalance_n

    # set iterative queue rebalance algorithm
    ncd_iq_rebal = args.rebalance_iq

    # set rebalance method.
    if ncd_iq_rebal:
        rebalance_dryrun = dataif.rebalance_dryrun_by_iq
    else:
        # round robin logic to rebalance.
        rebalance_dryrun = dataif.rebalance_dryrun_by_cyc

        # restrict only one dry run for cycles based mode.
        ncd_rebal_n = 1

    # set check point to call rebalance in vswitch
    rctx = RebalContext
    rctx.rebal_tick_n = ncd_rebal_interval / ncd_sample_interval

    if ncd_rebal:
        # adjust length of the samples counter
        config.ncd_samples_max = min(
            ncd_rebal_interval / ncd_sample_interval, config.ncd_samples_max)

        rctx.rebal_mode = True
        rctx.rebal_quick = False

    config.ncd_samples_max = int(config.ncd_samples_max)

    # set signal handler to abort ncd
    signal.signal(signal.SIGINT, ncd_kill)
    signal.signal(signal.SIGTERM, ncd_kill)

    tctx = TraceContext
    if ncd_trace:
        tctx.trace_mode = True

    # start ctld thread to monitor control command and dispatch
    # necessary action.
    shutdown_event = threading.Event()
    tobj = CtlDThread(shutdown_event)
    tobj.daemon = True

    try:
        tobj.start()
    except threading.ThreadError:
        nlog.info("failed to start ctld thread ..")
        sys.exit(1)

    prev_var = 0
    cur_var = 0
    ncd_samples_max = config.ncd_samples_max
    min_sample_i = 0

    # retrieve ncd dump if exists.
    ncd_dump = None
    if os.path.exists(config.ncd_dump_file):
        nlog.info("reading previous configuration and data from %s" %
                  config.ncd_dump_file)
        with open(config.ncd_dump_file, 'r') as f:
            ncd_dump = json.load(f)

        ctx.events = ncd_dump["events"]
        tctx.trace_mode = ncd_dump["config"]["trace_mode"]
        rctx.rebal_mode = ncd_dump["config"]["rebalance_mode"]
        fh = ctx.log_handler
        if ncd_dump["config"]["verbose_log"]:
            fh.setLevel(logging.DEBUG)

    else:
        os.makedirs(os.path.dirname(config.ncd_dump_file), exist_ok=True)
        with open(config.ncd_dump_file, 'w') as f:
            json.dump({
                "time": "%s" % datetime.now(),
                "version": netcontrold.__version__,
                "events": ctx.events,
                "config": {
                    "trace_mode": tctx.trace_mode,
                    "rebalance_mode": rctx.rebal_mode,
                    "verbose_log": fh.level == logging.DEBUG,
                },
            }, f)

    # begin rebalance dry run
    while (1):
        try:
            collect_data(ncd_samples_max, ncd_sample_interval)
            min_sample_i += ncd_samples_max

            nlog.info("current pmd load:")
            for pmd_id in sorted(pmd_map.keys()):
                pmd = pmd_map[pmd_id]
                nlog.info("pmd id %d load %d" % (pmd_id, pmd.pmd_load))

            cur_var = dataif.pmd_load_variance(pmd_map)
            nlog.info("current pmd load variance: %d" % cur_var)

            # do not trace if rebalance dry-run in progress.
            if tctx.trace_mode:
                pmd_cb_list = []
                for pname in sorted(ctx.port_to_cls.keys()):
                    port = ctx.port_to_cls[pname]
                    drop = dataif.port_drop_ppm(port)
                    drop_min = config.ncd_cb_pktdrop_min
                    tx_retry = dataif.port_tx_retry(port)
                    do_cb = False
                    if drop[0] > drop_min:
                        nlog.info("port %s drop_rx %d ppm above %d ppm" %
                                  (port.name, drop[0], drop_min))
                        ctx.events.append((port.name, "rx_drop", ctx.last_ts))
                        do_cb = True

                    if drop[1] > drop_min:
                        nlog.info("port %s drop_tx %d ppm above %d ppm" %
                                  (port.name, drop[1], drop_min))
                        ctx.events.append((port.name, "tx_drop", ctx.last_ts))
                        do_cb = True

                    if tx_retry > config.ncd_samples_max:
                        nlog.info("port %s tx_retry %d above %d" %
                                  (port.name, tx_retry,
                                   config.ncd_samples_max))
                        ctx.events.append((port.name, "tx_retry", ctx.last_ts))
                        do_cb = True

                    if not do_cb:
                        # no pmd needs to be traceged.
                        continue

                    for pmd_id in sorted(pmd_map.keys()):
                        pmd = pmd_map[pmd_id]
                        if (pmd.find_port_by_name(port.name)):
                            pmd_cb_list.insert(0, pmd_id)

                if (len(pmd_cb_list) > 0):
                    pmds = " ".join(list(map(str, set(pmd_cb_list))))
                    cmd = "%s %s" % (ncd_trace_cb, pmds)
                    nlog.info("executing callback %s" % cmd)
                    data = util.exec_host_command(cmd)
                    nlog.info(data)

            if not rctx.rebal_mode:
                continue

            # At the minimum for deriving current load on pmds, all of
            # the sampling counters (of size config.ncd_samples_max) have
            # to be filled "every time" before other evaluations done.
            #
            # However, for quick rebalance, we fill all the counters
            # once, and then keep rolling with one counter across old
            # stats so that, we reduce time to sample before kicking off
            # rebalance (from sampling config.ncd_samples_max counters
            # to only one.
            #
            # As ovs internally refers all its 6 sample counters for any
            # stats we query, it is absolutely fine we roll with one
            # new sample and retain old n-1 samples to check for current
            # state of pmd and rxqs.
            #
            if rctx.rebal_quick and (min_sample_i >= config.ncd_samples_max):
                ncd_samples_max = 1
                min_sample_i = (config.ncd_samples_max - 1)
            else:
                ncd_samples_max = config.ncd_samples_max

            # dry-run pmd rebalance.
            rebal_rxq_n = 0
            if pmd_map:
                for i in range(0, ncd_rebal_n):
                    n = 0
                    n = rebalance_dryrun(pmd_map)
                    if (n > 0):
                        rebal_rxq_n += n

            # restart sampling when no dry-run performed.
            if rebal_rxq_n == 0:
                nlog.info("no dryrun performed.")
                continue

            else:
                # compare previous and current state of pmds.
                prev_var = cur_var
                cur_var = dataif.pmd_load_variance(pmd_map)
                nlog.info("pmd load after dry run:")
                for pmd_id in sorted(pmd_map.keys()):
                    pmd = pmd_map[pmd_id]
                    nlog.info("pmd id %d load %d" % (pmd_id, pmd.pmd_load))

                nlog.info("pmd load variance: previous %d, after dry run %d" %
                          (prev_var, cur_var))

                if (cur_var < prev_var):
                    diff = (prev_var - cur_var) * 100 / prev_var
                    if diff > config.ncd_pmd_load_improve_min:
                        rctx.apply_rebal = True

                # check if balance state of all pmds is reached
                if rctx.apply_rebal:
                    # check if rebalance call needed really.
                    if (rctx.rebal_tick >= rctx.rebal_tick_n):
                        rctx.rebal_tick = 0
                        cmd = rebalance_switch(pmd_map)
                        ctx.events.append(("pmd", "rebalance", ctx.last_ts))
                        nlog.info(
                            "vswitch command for current optimization is: %s"
                            % cmd)
                        rctx.apply_rebal = False

                        if (util.exec_host_command(cmd) == 1):
                            nlog.info(
                                "problem running this command.. "
                                "check vswitch!")
                            now = datetime.now()
                            now_ts = now.strftime("%Y-%m-%d %H:%M:%S")
                            ctx.events.append(("switch", "error", now_ts))

                        # sleep for few seconds before thrashing current
                        # dry-run
                        nlog.info(
                            "waiting for %d seconds "
                            "before new dry runs begin.."
                            % config.ncd_vsw_wait_min)
                        time.sleep(config.ncd_vsw_wait_min)
                    else:
                        nlog.info("minimum rebalance interval not met!"
                                  " now at %d sec"
                                  % (rctx.rebal_tick * ncd_sample_interval))
                else:
                    nlog.info("no new optimization found ..")

                # reset collected data
                pmd_map.clear()
                ctx.port_to_cls.clear()
                ctx.port_to_id.clear()
                ncd_samples_max = config.ncd_samples_max
                min_sample_i = 0

                nlog.info("dry-run reset.")

        except error.NcdShutdownExc:
            nlog.info("Exiting NCD ..")
            tobj.ncd_shutdown.set()
            sys.exit(1)

    tobj.join()


if __name__ == "__main__":
    ncd_main(sys.argv[1:])
    sys.exit(0)



#
#  Copyright (c) 2020 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from unittest import TestCase
from unittest import mock
from netcontrold.lib import dataif
from netcontrold.lib import config

_FX_CPU_INFO = """
processor       : 0
core id         : 0
physical id     : 0

processor       : 1
core id         : 1
physical id     : 0

processor       : 2
core id         : 0
physical id     : 0

processor       : 3
core id         : 1
physical id     : 0
"""


class Test_pmd_load_variance_OnePmd(TestCase):

    pmd_map = dict()
    core_id = 0

    # setup test environment
    def setUp(self):
        # turn off limited info shown in assert failure for pmd object.
        self.maxDiff = None

        # a noop handler for debug info log.
        class NlogNoop(object):

            def debug(self, *args):
                None

        dataif.Context.nlog = NlogNoop()

        # create one pmd object.
        fx_pmd = dataif.Dataif_Pmd(self.core_id)

        # let it be in numa 0.
        fx_pmd.numa_id = 0

        # add some cpu consumption for this pmd.
        for i in range(0, config.ncd_samples_max):
            fx_pmd.idle_cpu_cyc[i] = (1000 + (100 * i))
            fx_pmd.proc_cpu_cyc[i] = (5000 + (500 * i))
            fx_pmd.rx_cyc[i] = (10000 + (100 * i))

        self.pmd_map[self.core_id] = fx_pmd
        return

    def test_one_rxq(self):
        # retrieve pmd object.
        pmd = self.pmd_map[self.core_id]

        # one dummy port is required for this test.
        port_name = 'virtport'

        # create port class of name 'virtport'.
        dataif.make_dataif_port(port_name)

        # add port object into pmd.
        fx_port = pmd.add_port(port_name)
        fx_port.numa_id = pmd.numa_id

        # add a dummy rxq into port.
        fx_rxq = fx_port.add_rxq(0)

        # add some cpu consumption for this rxq.
        for i in range(0, config.ncd_samples_max):
            fx_rxq.cpu_cyc[i] = (1000 + (100 * i))

        dataif.update_pmd_load(self.pmd_map)
        variance_value = dataif.pmd_load_variance(self.pmd_map)

        self.assertEqual(variance_value, 0)

        # del port object from pmd.
        pmd.del_port(port_name)

    # Test case:
    #   With one pmd thread handling few one single-queue ports
    def test_many_ports(self):
        # retrieve pmd object.
        pmd = self.pmd_map[self.core_id]

        # one few dummy ports required for this test.
        for port_name in ('virtport1', 'virtport2', 'virtport3'):

            # create port class of name 'virtport'.
            dataif.make_dataif_port(port_name)

            # add port object into pmd.
            fx_port = pmd.add_port(port_name)
            fx_port.numa_id = pmd.numa_id

            # add a dummy rxq into port.
            fx_rxq = fx_port.add_rxq(0)

            # add some cpu consumption for this rxq.
            for i in range(0, config.ncd_samples_max):
                fx_rxq.cpu_cyc[i] = (1000 + (100 * i))

        dataif.update_pmd_load(self.pmd_map)
        variance_value = dataif.pmd_load_variance(self.pmd_map)

        self.assertEqual(variance_value, 0)

        for port_name in ('virtport1', 'virtport2', 'virtport3'):
            pmd.del_port(port_name)

# Fixture:
#   create two pmd thread objects where in, each has one single-queued port.


def fx_2pmd_for_1rxq_each(testobj):
    # retrieve pmd object.
    pmd1 = testobj.pmd_map[testobj.core1_id]
    pmd2 = testobj.pmd_map[testobj.core2_id]

    # one dummy port is required for this test.
    port1_name = 'virtport1'
    port2_name = 'virtport2'

    # create port class of name 'virtport'.
    dataif.make_dataif_port(port1_name)
    dataif.make_dataif_port(port2_name)

    # add port object into pmd.
    fx_port1 = pmd1.add_port(port1_name)
    fx_port1.numa_id = pmd1.numa_id
    fx_port2 = pmd2.add_port(port2_name)
    fx_port2.numa_id = pmd2.numa_id

    # add a dummy rxq into port.
    fx_p1rxq = fx_port1.add_rxq(0)
    fx_p1rxq.pmd = pmd1
    fx_p2rxq = fx_port2.add_rxq(0)
    fx_p2rxq.pmd = pmd2

    # add some cpu consumption for these rxqs.
    for i in range(0, config.ncd_samples_max):
        fx_p1rxq.cpu_cyc[i] = (1000 + (100 * i))
        fx_p2rxq.cpu_cyc[i] = (2000 + (200 * i))


# Fixture:
#   Create two pmd thread objects where in, one pmd has two single-queued
#   ports, while the other is idle (without any port/rxq).
def fx_1pmd_for_2rxq(testobj):
    # retrieve pmd object.
    pmd1 = testobj.pmd_map[testobj.core1_id]

    # one dummy port is required for this test.
    port1_name = 'virtport1'
    port2_name = 'virtport2'

    # create port class of name 'virtport'.
    dataif.make_dataif_port(port1_name)
    dataif.make_dataif_port(port2_name)

    # add port object into pmd.
    fx_port1 = pmd1.add_port(port1_name)
    fx_port1.numa_id = pmd1.numa_id

    # add second port as well into pmd 1
    fx_port2 = pmd1.add_port(port2_name)
    fx_port2.numa_id = pmd1.numa_id

    # add a dummy rxq into port.
    fx_p1rxq = fx_port1.add_rxq(0)
    fx_p1rxq.pmd = pmd1
    fx_p2rxq = fx_port2.add_rxq(0)
    fx_p2rxq.pmd = pmd1

    # add some cpu consumption for these rxqs.
    for i in range(0, config.ncd_samples_max):
        fx_p1rxq.cpu_cyc[i] = (1000 + (100 * i))
        fx_p2rxq.cpu_cyc[i] = (2000 + (200 * i))


class Test_pmd_load_variance_TwoPmd(TestCase):
    """
    Test variance_value for one or more rxq handled by twp pmds.
    """

    pmd_map = dict()
    core1_id = 0
    core2_id = 1

    # setup test environment
    def setUp(self):
        # turn off limited info shown in assert failure for pmd object.
        self.maxDiff = None

        # a noop handler for debug info log.
        class NlogNoop(object):

            def info(self, *args):
                None

            def debug(self, *args):
                None

        dataif.Context.nlog = NlogNoop()

        # create one pmd object.
        fx_pmd1 = dataif.Dataif_Pmd(self.core1_id)
        fx_pmd2 = dataif.Dataif_Pmd(self.core2_id)

        # let it be in numa 0.
        fx_pmd1.numa_id = 0
        fx_pmd2.numa_id = 0

        # add some cpu consumption for these pmds.
        for i in range(0, config.ncd_samples_max):
            fx_pmd1.idle_cpu_cyc[i] = (1 + (1 * i))
            fx_pmd1.proc_cpu_cyc[i] = (900 + (90 * i))
            fx_pmd1.rx_cyc[i] = (1000 + (100 * i))

        for i in range(0, config.ncd_samples_max):
            fx_pmd2.idle_cpu_cyc[i] = (1000 + (100 * i))
            fx_pmd2.proc_cpu_cyc[i] = (9500 + (950 * i))
            fx_pmd2.rx_cyc[i] = (10000 + (1000 * i))

        self.pmd_map[self.core1_id] = fx_pmd1
        self.pmd_map[self.core2_id] = fx_pmd2
        return

    # Test case:
    #   With two threads from same numa, each handling only one single-queued
    #   port, variance_value
    def test_one_rxq_lnuma(self):
        # set different numa for pmds
        pmd1 = self.pmd_map[self.core1_id]
        pmd2 = self.pmd_map[self.core2_id]

        pmd1.numa_id = 0
        pmd2.numa_id = 0

        fx_2pmd_for_1rxq_each(self)

        dataif.update_pmd_load(self.pmd_map)
        variance_value = dataif.pmd_load_variance(self.pmd_map)

        self.assertEqual(int(variance_value), 17)

        # del port object from pmd.
        # TODO: create fx_ post deletion routine for clean up
        pmd1.del_port('virtport1')
        pmd2.del_port('virtport2')

    # Test case:
    #   With two threads from same numa, where one pmd thread is handling
    #   two single-queued ports, while the other is doing nothing,
    #   check variance_value.
    @mock.patch('netcontrold.lib.util.open')
    def test_two_rxq_lnuma(self, mock_open):
        mock_open.side_effect = [
            mock.mock_open(read_data=_FX_CPU_INFO).return_value
        ]

        # set different numa for pmds
        pmd1 = self.pmd_map[self.core1_id]
        pmd2 = self.pmd_map[self.core2_id]

        pmd1.numa_id = 0
        pmd2.numa_id = 0

        fx_1pmd_for_2rxq(self)

        dataif.update_pmd_load(self.pmd_map)
        variance_value = dataif.pmd_load_variance(self.pmd_map)

        self.assertEqual(int(variance_value), 17)

        # del port object from pmd.
        # TODO: create fx_ post deletion routine for clean up
        pmd1.del_port('virtport1')
        pmd1.del_port('virtport2')

    # Test case:
    #   With two threads from different numa, each handling one single-queued
    #   port, check variance_value.
    def test_one_rxq_rnuma(self):
        # set different numa for pmds
        pmd1 = self.pmd_map[self.core1_id]
        pmd2 = self.pmd_map[self.core2_id]

        pmd1.numa_id = 0
        pmd2.numa_id = 1

        fx_2pmd_for_1rxq_each(self)

        dataif.update_pmd_load(self.pmd_map)
        variance_value = dataif.pmd_load_variance(self.pmd_map)

        self.assertEqual(int(variance_value), 17)
        pmd1.del_port('virtport1')
        pmd2.del_port('virtport2')



#
#  Copyright (c) 2020 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from unittest import TestCase
from netcontrold.lib import dataif
from netcontrold.lib.error import ObjConsistencyExc


class TestDataif_add_rxq(TestCase):
    """
    Test for adding rxq to an existing port.
    """

    # setup test environment
    def setUp(self):
        # create one port object
        self.port1 = dataif.Port("port1")

        # add rxq to port object
        self.rxq2 = self.port1.add_rxq(2)

    # Test case:
    #   adding rxq to an existing port and checking whether
    #   it is added or not.
    def test_port_add_rxq_1(self):
        # add rxq to port object
        rxq1 = self.port1.add_rxq(1)

        self.assertEqual(self.port1.find_rxq_by_id(1), rxq1)

    # Test case:
    #   adding already existing rxq to port and checking whether
    #   addition is skipped.
    def test_port_add_rxq_2(self):
        self.assertRaises(ObjConsistencyExc, self.port1.add_rxq, 2)


class TestDataif_del_rxq(TestCase):
    """
    Test for deleting rxq from an existing port.
    """

    # setup test environment
    def setUp(self):
        # create one port object
        self.port1 = dataif.Port("port1")

        # add rxq1 to port object
        self.rxq1 = self.port1.add_rxq(1)

        # add rxq2 to port object
        self.rxq2 = self.port1.add_rxq(2)

    # Test case:
    #   deleting an existing rxq from a port and checking whether
    #   it is deleted or not.
    def test_port_del_rxq_1(self):
        # delete rxq1 from port object
        self.port1.del_rxq(1)

        self.assertEqual(self.port1.find_rxq_by_id(1), None)

    # Test case:
    #   deleting a non existing rxq from a port and checking whether
    #   deletion is skipped.
    def test_port_del_rxq_2(self):
        self.assertRaises(ObjConsistencyExc, self.port1.del_rxq, 3)


class TestDataif_find_rxq(TestCase):
    """
    Test for finding rxq from an existing port.
    """

    # setup test environment
    def setUp(self):
        # create one port object
        self.port1 = dataif.Port("port1")

        # add rxq1 to port object
        self.rxq1 = self.port1.add_rxq(1)

        # add rxq2 to port object
        self.rxq2 = self.port1.add_rxq(2)

    # Test case:
    #   finding an existing rxq from a port and checking whether
    #   it is found or not.
    def test_port_find_rxq_1(self):
        self.assertEqual(self.port1.find_rxq_by_id(1), self.rxq1)

    # Test case:
    #   finding a non existing rxq from a port and checking whether
    #   None is returned.
    def test_port_find_rxq_2(self):
        self.assertEqual(self.port1.find_rxq_by_id(3), None)


class TestDataif_add_port(TestCase):
    """
    Test for adding port to an existing pmd.
    """

    # setup for test environment
    def setUp(self):
        # create one pmd object
        self.pmd1 = dataif.Dataif_Pmd(1)

        # set pmd numa id
        self.pmd1.numa_id = 0

        dataif.make_dataif_port("port2")
        self.port2 = self.pmd1.add_port("port2")
        self.port2.numa_id = self.pmd1.numa_id

    # Test case:
    #   adding port to an existing pmd and checking whether
    #   it is added or not.
    def test_pmd_add_port_1(self):
        # create one port object
        dataif.make_dataif_port("port1")

        # add port object to pmd object
        port1 = self.pmd1.add_port("port1")

        # set port numa id
        port1.numa_id = self.pmd1.numa_id

        self.assertEqual(self.pmd1.find_port_by_name("port1"), port1)

    # Test case:
    #   adding existing port to a pmd and checking whether
    #   addition is skipped.
    def test_pmd_add_port_2(self):
        self.assertRaises(ObjConsistencyExc, self.pmd1.add_port, "port2")


class TestDataif_del_port(TestCase):
    """
    Test for deleting port from an existing pmd.
    """

    # setup for test environment
    def setUp(self):
        # create one pmd object
        self.pmd1 = dataif.Dataif_Pmd(1)

        # set pmd numa id
        self.pmd1.numa_id = 0

        # create port1 object
        dataif.make_dataif_port("port1")

        # add port object to pmd object
        self.port1 = self.pmd1.add_port("port1")

        # set port numa id
        self.port1.numa_id = self.pmd1.numa_id

        # create port3 object
        dataif.make_dataif_port("port3")

        # add port object to pmd object
        self.port3 = self.pmd1.add_port("port3")

        # set port numa id
        self.port3.numa_id = self.pmd1.numa_id

    # Test case:
    #   deleting an existing port from a pmd and checking whether
    #   it is deleted or not.
    def test_pmd_del_port_1(self):
        # delete port1 from pmd object
        self.pmd1.del_port("port1")

        self.assertEqual(self.pmd1.find_port_by_name("port1"), None)

    # Test case:
    #   deleting a non existing port from a pmd and checking whether
    #   deletion is skipped.
    def test_pmd_del_port_2(self):
        self.assertRaises(ObjConsistencyExc, self.pmd1.del_port, "port2")


class TestDataif_find_port(TestCase):
    """
    Test for finding port from an existing pmd.
    """

    # setup for test environment
    def setUp(self):
        # create one pmd object
        self.pmd1 = dataif.Dataif_Pmd(1)

        # set pmd numa id
        self.pmd1.numa_id = 0

        # create port1 object
        dataif.make_dataif_port("port1")

        # add port object to pmd object
        self.port1 = self.pmd1.add_port("port1")

        # set port numa id
        self.port1.numa_id = self.pmd1.numa_id

        # create port3 object
        dataif.make_dataif_port("port3")

        # add port object to pmd object
        self.port3 = self.pmd1.add_port("port3")

        # set port numa id
        self.port3.numa_id = self.pmd1.numa_id

    # Test case:
    #   finding an existing port from a pmd and checking whether
    #   it is found or not.
    def test_pmd_find_port_1(self):
        self.assertEqual(self.pmd1.find_port_by_name("port1"), self.port1)

    # Test case:
    #   finding a non existing port from a pmd and checking whether
    #   None is returned.
    def test_pmd_find_port_2(self):
        self.assertEqual(self.pmd1.find_port_by_name("port2"), None)


class TestDataif_count_rxq_empty(TestCase):
    """
    Test for counting rxq from an empty pmd.
    """

    # setup for test environment
    def setUp(self):
        # create one pmd object
        self.pmd1 = dataif.Dataif_Pmd(1)

        # set pmd numa id
        self.pmd1.numa_id = 0

        # create port1 object
        dataif.make_dataif_port("port1")

        # add port object to pmd object
        self.port1 = self.pmd1.add_port("port1")

        # set port numa id
        self.port1.numa_id = self.pmd1.numa_id

    # Test case:
    #   counting rxq from an empty pmd and checking whether
    #   no rxq is found.
    def test_pmd_count_rxq_1(self):
        self.assertEqual(self.pmd1.count_rxq(), 0)


class TestDataif_count_rxq(TestCase):
    """
    Test for counting rxq from an non empty pmd.
    """

    def setUp(self):
        # create one pmd object
        self.pmd1 = dataif.Dataif_Pmd(1)

        # set pmd numa id
        self.pmd1.numa_id = 0

        # create port1 object
        dataif.make_dataif_port("port1")

        # add port object to pmd object
        self.port1 = self.pmd1.add_port("port1")

        # set port numa id
        self.port1.numa_id = self.pmd1.numa_id

        # create port2 object
        dataif.make_dataif_port("port2")

        # add port object to pmd object
        self.port2 = self.pmd1.add_port("port2")

        # set port numa id
        self.port2.numa_id = self.pmd1.numa_id

        # add rxq1 to port1 object
        self.rxq1 = self.port1.add_rxq(1)

        # add rxq2 to port1 object
        self.rxq2 = self.port1.add_rxq(2)

        # add rxq1 to port2 object
        self.rxq1 = self.port2.add_rxq(1)

    # Test case:
    #   counting rxq from a non empty pmd and checking whether
    #   count is correct or not.
    def test_pmd_count_rxq_2(self):
        self.assertEqual(self.pmd1.count_rxq(), 3)

    # Test case:
    #   counting rxq from a non empty pmd after adding new rxq and checking
    #   whether count is correct.
    def test_pmd_count_rxq_3(self):
        # add rxq3 to port1 object
        _ = self.port1.add_rxq(3)

        # add rxq2 to port2 object
        _ = self.port2.add_rxq(2)

        self.assertEqual(self.pmd1.count_rxq(), 5)

    # Test case:
    #   counting rxq from a non empty pmd after deletion of rxq and checking
    #   whether count is correct
    def test_pmd_count_rxq_4(self):
        # delete rxq2 from port1 object
        self.port1.del_rxq(2)

        self.assertEqual(self.pmd1.count_rxq(), 2)



#
#  Copyright (c) 2020 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import copy
import pytest
from unittest import TestCase
from unittest import mock
from netcontrold.lib import dataif
from netcontrold.lib import config
from netcontrold.lib import util

# A noop handler for netcontrold logging.


class NlogNoop(object):

    def info(self, *args):
        prefix = "%s> " % (self.__class__.__name__)
        print("%s %s" % (prefix, "".join(args)))

    def debug(self, *args):
        prefix = "%s> " % (self.__class__.__name__)
        print("%s %s" % (prefix, "".join(args)))


_FX_CPU_INFO = """
processor       : 0
core id         : 0
physical id     : 0

processor       : 1
core id         : 1
physical id     : 0

processor       : 2
core id         : 0
physical id     : 0

processor       : 3
core id         : 1
physical id     : 0
"""

_FX_4X2CPU_INFO = """
processor       : 0
core id         : 0
physical id     : 0

processor       : 1
core id         : 1
physical id     : 0

processor       : 2
core id         : 0
physical id     : 1

processor       : 3
core id         : 1
physical id     : 1

processor       : 4
core id         : 0
physical id     : 0

processor       : 5
core id         : 1
physical id     : 0

processor       : 6
core id         : 0
physical id     : 1

processor       : 7
core id         : 1
physical id     : 1
"""

# -----------------------------------------------------------------------------
# Guidelines for setting pmd and rxq countes:
# -----------------------------------------------------------------------------
#
# rxq.cpu_cyc:
#   Number of cpu cycles consumed by a rxq during sample period.
#
# rxq.rx_cyc:
#   Number of packets received by a rxq during sample period.
#
# pmd.proc_cpu_cyc:
#   Number of cpu cycles consumed by a pmd since its stats start/reset.
#
# pmd.idle_cpu_cyc:
#   Number of cpu cycles not consumed by a pmd since its stats start/reset.
#
# pmd.rx_cyc:
#   Number of packets received by a pmd since its stats start/reset.
#
# Testobj.setup()
#   In setting up test environment, one or more pmd could be created for the
#   tests in that suite to use. By default, setup() function should create
#   a pmd for its load capacity at 96 % (i.e 1 % more than rebalance
#   threshold).
#
#   pmd.proc_cpu_cyc = 96
#   pmd.idle_cpu_cyc = 04
#   pmd.rx_cyc = 96
#
#   Hence that pmd would be shown at 96 % load capacity. When there are more
#   than one pmd to be created, at the minimum setting up one pmd to exceed
#   rebalance threshold value is must. The other pmds should setup load
#   capacity for 90 % in similar counter settings as above.
#
#   pmd.proc_cpu_cyc = 90
#   pmd.idle_cpu_cyc = 10
#   pmd.rx_cyc = 90
#
#   In all above, rx_cyc of a pmd should always reflect at proc_cpu_cyc
#   of that pmd for an easier calculation (i.e one packet per cpu cyc).
#
#   One or more rxq participating in a pmd for its load capacity, should
#   split proc_cpu_cyc among themself for a suitable rxq consumption level.
#
#   For eg, rxq1 and rxq2 in a pmd of load level 70 % could have
#   rxq1.cpu_cyc = 50
#   rxq1.rx_cyc = 50
#   rxq2.cpu_cyc = 20
#   rxq2.rx_cyc = 20
#
#   Every test should modify the default values provided by setup() function
#   for that test suite or fx_ function which was for a common usage across
#   the tests.
# -----------------------------------------------------------------------------


class TestRebalDryrun_OnePmd(TestCase):
    """
    Test rebalance for one or more rxq handled by one pmd.
    """

    rebalance_dryrun = dataif.rebalance_dryrun_by_cyc
    pmd_map = dict()
    core_id = 0

    # setup test environment
    def setUp(self):
        util.Memoize.forgot = True

        # turn off limited info shown in assert failure for pmd object.
        self.maxDiff = None

        dataif.Context.nlog = NlogNoop()

        # create one pmd object.
        fx_pmd = dataif.Dataif_Pmd(self.core_id)

        # let it be in numa 0.
        fx_pmd.numa_id = 0

        # add some cpu consumption for this pmd.
        for i in range(0, config.ncd_samples_max):
            fx_pmd.idle_cpu_cyc[i] = (4 * (i + 1))
            fx_pmd.proc_cpu_cyc[i] = (96 * (i + 1))
            fx_pmd.rx_cyc[i] = (96 * (i + 1))

        fx_pmd.cyc_idx = config.ncd_samples_max - 1
        self.pmd_map[self.core_id] = fx_pmd
        return

    # Test case:
    #   With one pmd thread handling only one single-queue port, check whether
    #   rebalance is skipped.
    def test_one_rxq(self):
        # retrieve pmd object.
        pmd = self.pmd_map[self.core_id]

        # one dummy port is required for this test.
        port_name = 'virtport'

        # create port class of name 'virtport'.
        dataif.make_dataif_port(port_name)

        # add port object into pmd.
        fx_port = pmd.add_port(port_name)
        fx_port.numa_id = pmd.numa_id

        # add a dummy rxq into port.
        fx_rxq = fx_port.add_rxq(0)

        # add some cpu consumption for this rxq.
        for i in range(0, config.ncd_samples_max):
            fx_rxq.cpu_cyc[i] = 96
            fx_rxq.rx_cyc[i] = 96

        # test dryrun
        n_reb_rxq = type(self).rebalance_dryrun(self.pmd_map)

        # validate results
        # 1. no rxqs be rebalanced.
        self.assertEqual(n_reb_rxq, -1)
        # 2. check pmd load
        self.assertEqual(dataif.pmd_load(pmd), 96.0)

        # del port object from pmd.
        pmd.del_port(port_name)

    # Test case:
    #   With one pmd thread handling few one single-queue ports, check whether
    #   rebalance is skipped.
    def test_many_rxq(self):
        # retrieve pmd object.
        pmd = self.pmd_map[self.core_id]

        # one few dummy ports required for this test.
        for port_name in ('virtport1', 'virtport2', 'virtport3'):

            # create port class of name 'virtport'.
            dataif.make_dataif_port(port_name)

            # add port object into pmd.
            fx_port = pmd.add_port(port_name)
            fx_port.numa_id = pmd.numa_id

            # add a dummy rxq into port.
            fx_rxq = fx_port.add_rxq(0)

            # add some cpu consumption for this rxq.
            for i in range(0, config.ncd_samples_max):
                fx_rxq.cpu_cyc[i] = 32
                fx_rxq.rx_cyc[i] = 32

        # test dryrun
        n_reb_rxq = type(self).rebalance_dryrun(self.pmd_map)

        # validate results
        # 1. no rxqs be rebalanced.
        self.assertEqual(n_reb_rxq, -1)
        # 2. check pmd load
        self.assertEqual(dataif.pmd_load(pmd), 96.0)

        # del port object from pmd.
        for port_name in ('virtport1', 'virtport2', 'virtport3'):
            pmd.del_port(port_name)


# Fixture:
#   create two pmd thread objects where in, each has one single-queued port.
def fx_2pmd_each_1rxq(testobj):
    # retrieve pmd object.
    pmd1 = testobj.pmd_map[testobj.core1_id]
    pmd2 = testobj.pmd_map[testobj.core2_id]

    # one dummy port is required for this test.
    port1_name = 'virtport1'
    port2_name = 'virtport2'

    # create port class of name 'virtport'.
    dataif.make_dataif_port(port1_name)
    dataif.make_dataif_port(port2_name)

    # add port object into pmd.
    fx_port1 = pmd1.add_port(port1_name)
    fx_port1.numa_id = pmd1.numa_id
    fx_port2 = pmd2.add_port(port2_name)
    fx_port2.numa_id = pmd2.numa_id

    # add a dummy rxq into port.
    fx_p1rxq = fx_port1.add_rxq(0)
    fx_p1rxq.pmd = pmd1
    fx_p2rxq = fx_port2.add_rxq(0)
    fx_p2rxq.pmd = pmd2

    # add some cpu consumption for these rxqs.
    for i in range(0, config.ncd_samples_max):
        fx_p1rxq.cpu_cyc[i] = 96
        fx_p1rxq.rx_cyc[i] = 96
        fx_p2rxq.cpu_cyc[i] = 90
        fx_p2rxq.rx_cyc[i] = 90


# Fixture:
#   Create two pmd thread objects where in, one pmd has two single-queued
#   ports, while the other is idle (without any port/rxq).
def fx_2pmd_one_empty(testobj):
    # retrieve pmd object.
    pmd1 = testobj.pmd_map[testobj.core1_id]

    # one dummy port is required for this test.
    port1_name = 'virtport1'
    port2_name = 'virtport2'

    # create port class of name 'virtport'.
    dataif.make_dataif_port(port1_name)
    dataif.make_dataif_port(port2_name)

    # add port object into pmd.
    fx_port1 = pmd1.add_port(port1_name)
    fx_port1.numa_id = pmd1.numa_id

    # add second port as well into pmd 1 for imbalance.
    fx_port2 = pmd1.add_port(port2_name)
    fx_port2.numa_id = pmd1.numa_id

    # add a dummy rxq into port.
    fx_p1rxq = fx_port1.add_rxq(0)
    fx_p1rxq.pmd = pmd1
    fx_p2rxq = fx_port2.add_rxq(0)
    fx_p2rxq.pmd = pmd1

    # add some cpu consumption for these rxqs.
    for i in range(0, config.ncd_samples_max):
        fx_p1rxq.cpu_cyc[i] = 6
        fx_p1rxq.rx_cyc[i] = 6
        fx_p2rxq.cpu_cyc[i] = 90
        fx_p2rxq.rx_cyc[i] = 90


# Fixture:
#   Create two pmd thread objects where in, each pmd has two single-queued
#   ports.
def fx_2pmd_each_2rxq(testobj):
    # retrieve pmd object.
    pmd1 = testobj.pmd_map[testobj.core1_id]
    pmd2 = testobj.pmd_map[testobj.core2_id]

    # one dummy port is required for this test.
    port1_name = 'virtport1'
    port2_name = 'virtport2'
    port3_name = 'virtport3'
    port4_name = 'virtport4'

    # create port class of name 'virtport'.
    dataif.make_dataif_port(port1_name)
    dataif.make_dataif_port(port2_name)
    dataif.make_dataif_port(port3_name)
    dataif.make_dataif_port(port4_name)

    # add port object into pmd.
    fx_port1 = pmd1.add_port(port1_name)
    fx_port1.numa_id = pmd1.numa_id
    fx_port2 = pmd2.add_port(port2_name)
    fx_port2.numa_id = pmd2.numa_id
    fx_port3 = pmd1.add_port(port3_name)
    fx_port3.numa_id = pmd1.numa_id
    fx_port4 = pmd2.add_port(port4_name)
    fx_port4.numa_id = pmd2.numa_id

    # add a dummy rxq into port.
    fx_p1rxq = fx_port1.add_rxq(0)
    fx_p1rxq.pmd = pmd1
    fx_p2rxq = fx_port2.add_rxq(0)
    fx_p2rxq.pmd = pmd2
    fx_p3rxq = fx_port3.add_rxq(0)
    fx_p3rxq.pmd = pmd1
    fx_p4rxq = fx_port4.add_rxq(0)
    fx_p4rxq.pmd = pmd2

    # add some cpu consumption for these rxqs.
    # order of rxqs based on cpu consumption: rxqp1,rxqp2,rxqp3,rxqp4
    for i in range(0, config.ncd_samples_max):
        fx_p1rxq.cpu_cyc[i] = 70
        fx_p1rxq.rx_cyc[i] = 70
        fx_p2rxq.cpu_cyc[i] = 65
        fx_p2rxq.rx_cyc[i] = 65
        fx_p3rxq.cpu_cyc[i] = 26
        fx_p3rxq.rx_cyc[i] = 26
        fx_p4rxq.cpu_cyc[i] = 25
        fx_p4rxq.rx_cyc[i] = 25

# Fixture:
#   Create two pmd thread objects where in, two queued ports split
#   among pmds.


def fx_2pmd_each_1p2rxq(testobj):
    # retrieve pmd object.
    pmd1 = testobj.pmd_map[testobj.core1_id]
    pmd2 = testobj.pmd_map[testobj.core2_id]

    # dummy ports required for this test.
    port1_name = 'virtport1'
    port2_name = 'virtport2'

    # create port class of name 'virtport'.
    dataif.make_dataif_port(port1_name)
    dataif.make_dataif_port(port2_name)

    # add port object into pmd.
    fx_port11 = pmd1.add_port(port1_name)
    fx_port11.numa_id = pmd1.numa_id
    fx_port22 = pmd2.add_port(port2_name)
    fx_port22.numa_id = pmd2.numa_id
    fx_port21 = pmd1.add_port(port2_name)
    fx_port21.numa_id = pmd1.numa_id
    fx_port12 = pmd2.add_port(port1_name)
    fx_port12.numa_id = pmd2.numa_id

    # add a dummy rxq into port.
    fx_p1rxq1 = fx_port11.add_rxq(0)
    fx_p1rxq1.pmd = pmd1
    fx_p2rxq2 = fx_port22.add_rxq(1)
    fx_p2rxq2.pmd = pmd2
    fx_p2rxq1 = fx_port21.add_rxq(0)
    fx_p2rxq1.pmd = pmd1
    fx_p1rxq2 = fx_port12.add_rxq(1)
    fx_p1rxq2.pmd = pmd2

    # add some cpu consumption for these rxqs.
    # order of rxqs based on cpu consumption: rxq1p1,rxq2p2,rxq1p2,rxq2p1
    for i in range(0, config.ncd_samples_max):
        fx_p1rxq1.cpu_cyc[i] = 90
        fx_p1rxq1.rx_cyc[i] = 90
        fx_p2rxq2.cpu_cyc[i] = 60
        fx_p2rxq2.rx_cyc[i] = 60
        fx_p1rxq2.cpu_cyc[i] = 30
        fx_p1rxq2.rx_cyc[i] = 30
        fx_p2rxq1.cpu_cyc[i] = 6
        fx_p2rxq1.rx_cyc[i] = 6


class TestRebalDryrun_TwoPmd(TestCase):
    """
    Test rebalance for one or more rxq handled by two pmds.
    """

    rebalance_dryrun = dataif.rebalance_dryrun_by_cyc
    pmd_map = dict()
    core1_id = 0
    core2_id = 1

    # setup test environment
    def setUp(self):
        util.Memoize.forgot = True

        # turn off limited info shown in assert failure for pmd object.
        self.maxDiff = None

        dataif.Context.nlog = NlogNoop()

        # create one pmd object.
        fx_pmd1 = dataif.Dataif_Pmd(self.core1_id)
        fx_pmd2 = dataif.Dataif_Pmd(self.core2_id)

        # let it be in numa 0.
        fx_pmd1.numa_id = 0
        fx_pmd2.numa_id = 0

        # add some cpu consumption for these pmds.
        for i in range(0, config.ncd_samples_max):
            fx_pmd1.idle_cpu_cyc[i] = (4 * (i + 1))
            fx_pmd1.proc_cpu_cyc[i] = (96 * (i + 1))
            fx_pmd1.rx_cyc[i] = (96 * (i + 1))

        fx_pmd1.cyc_idx = config.ncd_samples_max - 1

        for i in range(0, config.ncd_samples_max):
            fx_pmd2.idle_cpu_cyc[i] = (10 * (i + 1))
            fx_pmd2.proc_cpu_cyc[i] = (90 * (i + 1))
            fx_pmd2.rx_cyc[i] = (90 * (i + 1))

        fx_pmd2.cyc_idx = config.ncd_samples_max - 1

        self.pmd_map[self.core1_id] = fx_pmd1
        self.pmd_map[self.core2_id] = fx_pmd2
        return

    # Test case:
    #   With two threads from same numa, each handling only one single-queued
    #   port, check whether rebalance is skipped.
    def test_one_rxq_lnuma(self):
        # set same numa for pmds
        pmd1 = self.pmd_map[self.core1_id]
        pmd2 = self.pmd_map[self.core2_id]
        pmd1.numa_id = 0
        pmd2.numa_id = 0

        # create rxq
        fx_2pmd_each_1rxq(self)

        # update pmd load values
        dataif.update_pmd_load(self.pmd_map)

        # copy original pmd objects
        pmd_map = copy.deepcopy(self.pmd_map)

        # test dryrun
        n_reb_rxq = type(self).rebalance_dryrun(self.pmd_map)

        # validate results
        # 1. all two rxqs be rebalanced.
        self.assertEqual(n_reb_rxq, -1, "no rebalance expected")
        # 2. each pmd is not updated.
        self.assertEqual(pmd_map[self.core1_id], pmd1)
        self.assertEqual(pmd_map[self.core2_id], pmd2)
        # 3. check pmd load
        self.assertEqual(dataif.pmd_load(pmd1), 96.0)
        self.assertEqual(dataif.pmd_load(pmd2), 90.0)

        # del port object from pmd.
        # TODO: create fx_ post deletion routine for clean up
        pmd1.del_port('virtport1')
        pmd2.del_port('virtport2')

    # Test case:
    #   With two threads from same numa, where one pmd thread is handling
    #   two single-queued ports, while the other pmd is empty,
    #   check whether rebalance is performed.
    #   Scope is to check if only one rxq is moved to empty pmd.
    #
    #   order of rxqs based on cpu consumption: rxqp2,rxqp1
    #   order of pmds for rebalance dryrun: pmd1,pmd2
    #
    #   1. rxqp2(pmd1) -NOREB-> rxqp2(pmd1)
    #      rxqp1(pmd1)
    #        -  (pmd2)
    #
    #   2. rxqp2(pmd1) -NOREB-> rxqp2(pmd1)
    #      rxqp1(pmd1) --+--+-> rxqp1(reb_pmd2)
    #
    @mock.patch('netcontrold.lib.util.open')
    def test_two_1rxq_with_empty_lnuma(self, mock_open):
        mock_open.side_effect = [
            mock.mock_open(read_data=_FX_CPU_INFO).return_value
        ]

        # set same numa for pmds
        pmd1 = self.pmd_map[self.core1_id]
        pmd2 = self.pmd_map[self.core2_id]
        pmd1.numa_id = 0
        pmd2.numa_id = 0

        # let pmd2 be idle
        for i in range(0, config.ncd_samples_max):
            pmd2.idle_cpu_cyc[i] = (100 * (i + 1))
            pmd2.proc_cpu_cyc[i] = (0 * (i + 1))
            pmd2.rx_cyc[i] = (0 * (i + 1))

        # create rxq
        fx_2pmd_one_empty(self)

        # update pmd load values
        dataif.update_pmd_load(self.pmd_map)

        # copy original pmd objects
        pmd_map = copy.deepcopy(self.pmd_map)

        # test dryrun
        n_reb_rxq = type(self).rebalance_dryrun(self.pmd_map)

        # validate results
        # 1. all two rxqs be rebalanced.
        self.assertEqual(n_reb_rxq, 1, "one rxq to be rebalanced")
        # 2. each pmd is updated.
        self.assertNotEqual(pmd_map[self.core1_id], pmd1)
        self.assertNotEqual(pmd_map[self.core2_id], pmd2)
        # 3. check rxq map after dryrun.
        port1 = pmd1.find_port_by_name('virtport1')
        port2 = pmd1.find_port_by_name('virtport2')
        # 3.a rxqp2 remains in pmd1
        self.assertEqual(port2.rxq_rebalanced, {})
        self.assertEqual(port2.find_rxq_by_id(0).pmd.id, pmd1.id)
        # 3.a rxqp1 moves into pmd2
        self.assertEqual(port1.rxq_rebalanced[0], pmd2.id)
        # 4. check pmd load
        self.assertEqual(dataif.pmd_load(pmd1), 90.0)
        self.assertEqual(dataif.pmd_load(pmd2), 6.0)

        # del port object from pmd.
        # TODO: create fx_ post deletion routine for clean up
        pmd1.del_port('virtport1')
        pmd1.del_port('virtport2')

    # Test case:
    #   With two threads from same numa, where each pmd thread is  handling
    #   one queue from two-queued ports. check whether rebalance is performed.
    #   Scope is to check if rxq affinity is retained.
    #
    #   order of rxqs based on cpu consumption: rxq1p1,rxq2p2,rxq2p1,rxq1p2
    #   order of pmds for rebalance dryrun: pmd1,pmd2,pmd2,pmd1
    #
    #   1. rxq1p1(pmd1) 90% -NOREB-> rxq1p1(pmd1)
    #      rxq1p2(pmd1)  6% -NOREB-> rxq1p2(pmd1)
    #      rxq2p2(pmd2) 60% -NOREB-> rxq2p2(pmd2)
    #      rxq2p1(pmd2) 30% -NOREB-> rxq2p1(pmd1)
    #
    @mock.patch('netcontrold.lib.util.open')
    def test_two_1p2rxq_lnuma_norb(self, mock_open):
        mock_open.side_effect = [
            mock.mock_open(read_data=_FX_CPU_INFO).return_value
        ]

        # set same numa for pmds
        pmd1 = self.pmd_map[self.core1_id]
        pmd2 = self.pmd_map[self.core2_id]
        pmd1.numa_id = 0
        pmd2.numa_id = 0

        # create rxq
        fx_2pmd_each_1p2rxq(self)

        # update pmd load values
        dataif.update_pmd_load(self.pmd_map)

        # copy original pmd objects
        pmd_map = copy.deepcopy(self.pmd_map)

        # test dryrun
        n_reb_rxq = type(self).rebalance_dryrun(self.pmd_map)

        # validate results
        # 1. all four rxqs be rebalanced.
        self.assertEqual(n_reb_rxq, 0, "no rebalance expected")
        # 2. each pmd is updated.
        self.assertEqual(pmd_map[self.core1_id], pmd1)
        self.assertEqual(pmd_map[self.core2_id], pmd2)
        # 3. check rxq map after dryrun.
        port11 = pmd1.find_port_by_name('virtport1')
        port12 = pmd2.find_port_by_name('virtport1')
        port21 = pmd1.find_port_by_name('virtport2')
        port22 = pmd2.find_port_by_name('virtport2')
        self.assertEqual(port11.rxq_rebalanced, {})
        self.assertEqual(port22.rxq_rebalanced, {})
        self.assertEqual(port21.rxq_rebalanced, {})
        self.assertEqual(port12.rxq_rebalanced, {})
        # 4. check pmd load
        self.assertEqual(dataif.pmd_load(pmd1), 96.0)
        self.assertEqual(dataif.pmd_load(pmd2), 90.0)

        # del port object from pmd.
        # TODO: create fx_ post deletion routine for clean up
        pmd1.del_port('virtport1')
        pmd2.del_port('virtport2')

    # Test case:
    #   With two threads from same numa, where each pmd thread is  handling
    #   one queue from two-queued ports. check whether rebalance is performed.
    #   Scope is to check if rebalancing is not done on pmd that already
    #   has same port but different rxq.
    #   ( For now rebalance is allowed by switch, so follow)
    #
    #   order of rxqs based on cpu consumption: rxq2p2,rxq1p1,rxq1p3,
    #                                           rxq1p2,rxq2p1
    #   order of pmds for rebalance dryrun: pmd2,pmd1,pmd1,pmd2,pmd2
    #
    #   1. rxq1p1(pmd1) 66% -NOREB-> rxq1p1(pmd1)
    #      rxq1p3(pmd1) 22%
    #      rxq1p2(pmd1)  8%
    #      rxq2p2(pmd2) 86% -NOREB-> rxq2p2(pmd2)
    #      rxq2p1(pmd2)  4%
    #
    #   2. rxq2p2(pmd2) 86% -NOREB-> rxq2p2(pmd2)
    #      rxq1p1(pmd1) 66% -NOREB-> rxq1p1(pmd1)
    #      rxq1p3(pmd1) 22% -NOREB-> rxq1p3(pmd1)
    #      rxq1p2(pmd1)  8% --+--+-> rxq1p2(reb_pmd2)
    #      rxq2p1(pmd2)  4% -NOREB-> rxq2p1(pmd2)

    #
    @mock.patch('netcontrold.lib.util.open')
    def test_two_1p2rxq_lnuma(self, mock_open):
        mock_open.side_effect = [
            mock.mock_open(read_data=_FX_CPU_INFO).return_value
        ]

        # set same numa for pmds
        pmd1 = self.pmd_map[self.core1_id]
        pmd2 = self.pmd_map[self.core2_id]
        pmd1.numa_id = 0
        pmd2.numa_id = 0

        # create rxq
        fx_2pmd_each_1p2rxq(self)

        # we need an extra port to break retaining some ports.
        dataif.make_dataif_port('virtport3')
        port31 = pmd1.add_port('virtport3')
        port31.numa_id = pmd1.numa_id
        p3rxq1 = port31.add_rxq(0)
        p3rxq1.pmd = pmd1

        # update some cpu consumption for these rxqs.
        # order of rxqs based on cpu consumption:
        # rxq2p2,rxq1p1,rxq1p3,rxq1p2,rxq2p1
        port11 = pmd1.find_port_by_name('virtport1')
        port12 = pmd2.find_port_by_name('virtport1')
        port21 = pmd1.find_port_by_name('virtport2')
        port22 = pmd2.find_port_by_name('virtport2')
        p1rxq1 = port11.find_rxq_by_id(0)
        p1rxq2 = port12.find_rxq_by_id(1)
        p2rxq1 = port21.find_rxq_by_id(0)
        p2rxq2 = port22.find_rxq_by_id(1)

        for i in range(0, config.ncd_samples_max):
            p2rxq2.cpu_cyc[i] = 86
            p1rxq1.cpu_cyc[i] = 66
            p3rxq1.cpu_cyc[i] = 22
            p2rxq1.cpu_cyc[i] = 8
            p1rxq2.cpu_cyc[i] = 4

            p2rxq2.rx_cyc[i] = 86
            p1rxq1.rx_cyc[i] = 66
            p3rxq1.rx_cyc[i] = 22
            p2rxq1.rx_cyc[i] = 8
            p1rxq2.rx_cyc[i] = 4

        # update pmd load values
        dataif.update_pmd_load(self.pmd_map)

        # copy original pmd objects
        pmd_map = copy.deepcopy(self.pmd_map)

        # test dryrun
        n_reb_rxq = type(self).rebalance_dryrun(self.pmd_map)

        # validate results
        # 1. all four rxqs be rebalanced.
        self.assertEqual(n_reb_rxq, 1, "one rxq to be rebalanced")
        # 2. each pmd is updated.
        self.assertNotEqual(pmd_map[self.core1_id], pmd1)
        self.assertNotEqual(pmd_map[self.core2_id], pmd2)
        # 3. check rxq map after dryrun.
        self.assertEqual(port11.rxq_rebalanced, {})
        self.assertEqual(port22.rxq_rebalanced, {})
        self.assertEqual(port31.rxq_rebalanced, {})
        self.assertEqual(port21.rxq_rebalanced[0], pmd2.id)
        self.assertEqual(port12.rxq_rebalanced, {})
        # 4. check pmd load
        self.assertEqual(dataif.pmd_load(pmd1), 88.0)
        self.assertEqual(dataif.pmd_load(pmd2), 98.0)

        # del port object from pmd.
        # TODO: create fx_ post deletion routine for clean up
        pmd1.del_port('virtport1')
        pmd1.del_port('virtport3')
        pmd2.del_port('virtport2')

    # Test case:
    #   With two threads from same numa, where each pmd thread is handling
    #   two single-queued ports. check whether rebalance is performed.
    #   Scope is to check if rxq from a pmd which was a rebalancing pmd
    #   before, is assigned other pmd successfully.
    #
    #   order of rxqs based on cpu consumption: rxqp1,rxqp2,rxqp3,rxqp4
    #   order of pmds for rebalance dryrun: pmd1,pmd2,pmd2,pmd1
    #
    #   1. rxqp1(pmd1) -NOREB-> rxqp1(pmd1)
    #      rxqp2(pmd2) -NOREB-> rxqp2(pmd2)
    #      rxqp3(pmd1) --+--+-> rxqp3(reb_pmd2)
    #      rxqp4(pmd2)
    #
    #   2. rxqp1(pmd1) -NOREB-> rxqp1(pmd1)
    #      rxqp2(pmd2) -NOREB-> rxqp2(pmd2)
    #      rxqp3(pmd1) --+--+-> rxqp3(reb_pmd2)
    #      rxqp4(pmd2) --+--+-> rxqp4(reb_pmd1)
    #
    @mock.patch('netcontrold.lib.util.open')
    def test_four_1rxq_lnuma(self, mock_open):
        mock_open.side_effect = [
            mock.mock_open(read_data=_FX_CPU_INFO).return_value
        ]

        # set same numa for pmds
        pmd1 = self.pmd_map[self.core1_id]
        pmd2 = self.pmd_map[self.core2_id]
        pmd1.numa_id = 0
        pmd2.numa_id = 0

        # create rxq
        fx_2pmd_each_2rxq(self)

        # update pmd load values
        dataif.update_pmd_load(self.pmd_map)

        # copy original pmd objects
        pmd_map = copy.deepcopy(self.pmd_map)

        # test dryrun
        n_reb_rxq = type(self).rebalance_dryrun(self.pmd_map)

        # validate results
        # 1. all four rxqs be rebalanced.
        self.assertEqual(n_reb_rxq, 2, "two rxqs to be rebalanced")
        # 2. each pmd is updated.
        self.assertNotEqual(pmd_map[self.core1_id], pmd1)
        self.assertNotEqual(pmd_map[self.core2_id], pmd2)
        # 3. check rxq map after dryrun.
        port1 = pmd1.find_port_by_name('virtport1')
        port2 = pmd2.find_port_by_name('virtport2')
        port3 = pmd1.find_port_by_name('virtport3')
        port4 = pmd2.find_port_by_name('virtport4')
        self.assertEqual(port1.rxq_rebalanced, {})
        self.assertEqual(port2.rxq_rebalanced, {})
        self.assertEqual(port3.rxq_rebalanced[0], pmd2.id)
        self.assertEqual(port4.rxq_rebalanced[0], pmd1.id)
        # 4. check pmd load
        self.assertEqual(dataif.pmd_load(pmd1), (96.0 - 26.0 + 25.0))
        self.assertEqual(dataif.pmd_load(pmd2), (90.0 - 25.0 + 26.0))

        # del port object from pmd.
        # TODO: create fx_ post deletion routine for clean up
        pmd1.del_port('virtport1')
        pmd2.del_port('virtport2')

    # Test case:
    #   With two threads from same numa, where each pmd thread is handling
    #   two single-queued ports. Of them, only one rxq is busy while the
    #   rest are idle. check whether rebalance is not moving busy rxq
    #   from its pmd, while rest (which are idle rxqs) could be repinned
    #   accordingly.
    #
    #   order of rxqs based on cpu consumption: rxqp4 (and some order on
    #                                           rxqp2,rxqp3,rxqp4)
    #   order of pmds for rebalance dryrun: pmd1,pmd2,pmd2,pmd1
    #
    #   1. rxqp1(pmd1)
    #      rxqp2(pmd2)
    #      rxqp3(pmd1)
    #      rxqp4(pmd2) -NOREB-> rxqp4(pmd2)
    #
    @mock.patch('netcontrold.lib.util.open')
    def test_four_1rxq_skip_lnuma(self, mock_open):
        mock_open.side_effect = [
            mock.mock_open(read_data=_FX_CPU_INFO).return_value
        ]

        # set same numa for pmds
        pmd1 = self.pmd_map[self.core1_id]
        pmd2 = self.pmd_map[self.core2_id]
        pmd1.numa_id = 0
        pmd2.numa_id = 0

        # create rxq
        fx_2pmd_each_2rxq(self)

        # except one rxq, let rest be idle.
        port1 = pmd1.find_port_by_name('virtport1')
        port2 = pmd2.find_port_by_name('virtport2')
        port3 = pmd1.find_port_by_name('virtport3')
        port4 = pmd2.find_port_by_name('virtport4')
        p1rxq = port1.find_rxq_by_id(0)
        p2rxq = port2.find_rxq_by_id(0)
        p3rxq = port3.find_rxq_by_id(0)
        p4rxq = port4.find_rxq_by_id(0)
        for i in range(0, config.ncd_samples_max):
            p1rxq.cpu_cyc[i] = 0
            p2rxq.cpu_cyc[i] = 0
            p3rxq.cpu_cyc[i] = 0
            p4rxq.cpu_cyc[i] = 98

            p1rxq.rx_cyc[i] = 0
            p2rxq.rx_cyc[i] = 0
            p3rxq.rx_cyc[i] = 0
            p4rxq.rx_cyc[i] = 98

        # fix cpu consumption for these pmds.
        for i in range(0, config.ncd_samples_max):
            pmd1.idle_cpu_cyc[i] = (100 * (i + 1))
            pmd1.proc_cpu_cyc[i] = (0 + (0 * i))
            pmd1.rx_cyc[i] = (0 * (i + 1))

        for i in range(0, config.ncd_samples_max):
            pmd2.idle_cpu_cyc[i] = (2 * (i + 1))
            pmd2.proc_cpu_cyc[i] = (98 * (i + 1))
            pmd2.rx_cyc[i] = (98 * (i + 1))

        # update pmd load values
        dataif.update_pmd_load(self.pmd_map)

        # copy original pmd objects
        pmd_map = copy.deepcopy(self.pmd_map)

        # test dryrun
        n_reb_rxq = type(self).rebalance_dryrun(self.pmd_map)

        # validate results
        # 1. all four rxqs be rebalanced.
        self.assertEqual(n_reb_rxq, 0, "expected no rebalance")
        # 2. each pmd is not updated.
        self.assertEqual(pmd_map[self.core1_id], pmd1)
        self.assertEqual(pmd_map[self.core2_id], pmd2)
        # 3. check rxq map after dryrun.
        self.assertEqual(port1.rxq_rebalanced, {})
        self.assertEqual(port2.rxq_rebalanced, {})
        self.assertEqual(port3.rxq_rebalanced, {})
        self.assertEqual(port4.rxq_rebalanced, {})
        # 3.a and dry-run did not break original pinning.
        self.assertEqual(p4rxq.pmd.id, pmd2.id)
        # 4. check pmd load
        self.assertEqual(dataif.pmd_load(pmd1), 0)
        self.assertEqual(dataif.pmd_load(pmd2), 98.0)

        # del port object from pmd.
        # TODO: create fx_ post deletion routine for clean up
        pmd1.del_port('virtport1')
        pmd2.del_port('virtport2')
        pmd1.del_port('virtport3')
        pmd2.del_port('virtport4')

    # Test case:
    #   With two threads from same numa, where each pmd thread is handling
    #   two single-queued ports. Of them, all are busy. check whether
    #   rebalance is skipped.
    #
    #   order of rxqs based on cpu consumption: N/A
    #   order of pmds for rebalance dryrun: N/A
    #
    #   1. rxqp1(pmd1)
    #      rxqp2(pmd2)
    #      rxqp3(pmd1)
    #      rxqp4(pmd2)
    #
    @mock.patch('netcontrold.lib.util.open')
    def test_4busy_1rxq_skip_lnuma(self, mock_open):
        mock_open.side_effect = [
            mock.mock_open(read_data=_FX_CPU_INFO).return_value
        ]

        # set same numa for pmds
        pmd1 = self.pmd_map[self.core1_id]
        pmd2 = self.pmd_map[self.core2_id]
        pmd1.numa_id = 0
        pmd2.numa_id = 0

        # create rxq
        fx_2pmd_each_2rxq(self)

        # fix cpu consumption for these pmds.
        for i in range(0, config.ncd_samples_max):
            pmd1.idle_cpu_cyc[i] = (4 * (i + 1))
            pmd1.proc_cpu_cyc[i] = (96 * (i + 1))
            pmd1.rx_cyc[i] = (96 * (i + 1))

        for i in range(0, config.ncd_samples_max):
            pmd2.idle_cpu_cyc[i] = (4 * (i + 1))
            pmd2.proc_cpu_cyc[i] = (96 * (i + 1))
            pmd2.rx_cyc[i] = (96 * (i + 1))

        # all rxqs are busy
        port1 = pmd1.find_port_by_name('virtport1')
        port2 = pmd2.find_port_by_name('virtport2')
        port3 = pmd1.find_port_by_name('virtport3')
        port4 = pmd2.find_port_by_name('virtport4')
        p1rxq = port1.find_rxq_by_id(0)
        p2rxq = port2.find_rxq_by_id(0)
        p3rxq = port3.find_rxq_by_id(0)
        p4rxq = port4.find_rxq_by_id(0)
        for i in range(0, config.ncd_samples_max):
            p1rxq.cpu_cyc[i] = 70
            p2rxq.cpu_cyc[i] = 60
            p3rxq.cpu_cyc[i] = 26
            p4rxq.cpu_cyc[i] = 36

            p1rxq.rx_cyc[i] = 70
            p2rxq.rx_cyc[i] = 60
            p3rxq.rx_cyc[i] = 26
            p4rxq.rx_cyc[i] = 36

        # update pmd load values
        dataif.update_pmd_load(self.pmd_map)

        # copy original pmd objects
        pmd_map = copy.deepcopy(self.pmd_map)

        # test dryrun
        n_reb_rxq = type(self).rebalance_dryrun(self.pmd_map)

        # validate results
        port1 = pmd1.find_port_by_name('virtport1')
        port2 = pmd2.find_port_by_name('virtport2')
        port3 = pmd1.find_port_by_name('virtport3')
        port4 = pmd2.find_port_by_name('virtport4')
        # 1. all four rxqs be rebalanced.
        self.assertEqual(n_reb_rxq, -1, "skip rebalance")
        # 2. each pmd is not updated.
        self.assertEqual(pmd_map[self.core1_id], pmd1)
        self.assertEqual(pmd_map[self.core2_id], pmd2)
        # 3. check rxq map after dryrun.
        self.assertEqual(port1.rxq_rebalanced, {})
        self.assertEqual(port2.rxq_rebalanced, {})
        self.assertEqual(port3.rxq_rebalanced, {})
        self.assertEqual(port4.rxq_rebalanced, {})
        # 4. check pmd load
        self.assertEqual(dataif.pmd_load(pmd1), 96.0)
        self.assertEqual(dataif.pmd_load(pmd2), 96.0)

        # del port object from pmd.
        # TODO: create fx_ post deletion routine for clean up
        pmd1.del_port('virtport1')
        pmd2.del_port('virtport2')
        pmd1.del_port('virtport3')
        pmd2.del_port('virtport4')


# Fixture:
#   Create four pmd thread objects where in, each pmd has two single-queued
#   ports.


def fx_4pmd_each_2rxq(testobj):
    # retrieve pmd object.
    pmd1 = testobj.pmd_map[testobj.core1_id]
    pmd2 = testobj.pmd_map[testobj.core2_id]
    pmd3 = testobj.pmd_map[testobj.core3_id]
    pmd4 = testobj.pmd_map[testobj.core4_id]

    # one dummy port is required for this test.
    port1_name = 'virtport1'
    port2_name = 'virtport2'
    port3_name = 'virtport3'
    port4_name = 'virtport4'
    port5_name = 'virtport5'
    port6_name = 'virtport6'
    port7_name = 'virtport7'
    port8_name = 'virtport8'

    # create port class of name 'virtport'.
    dataif.make_dataif_port(port1_name)
    dataif.make_dataif_port(port2_name)
    dataif.make_dataif_port(port3_name)
    dataif.make_dataif_port(port4_name)
    dataif.make_dataif_port(port5_name)
    dataif.make_dataif_port(port6_name)
    dataif.make_dataif_port(port7_name)
    dataif.make_dataif_port(port8_name)

    # add port object into pmd.
    fx_port1 = pmd1.add_port(port1_name)
    fx_port1.numa_id = pmd1.numa_id
    fx_port2 = pmd2.add_port(port2_name)
    fx_port2.numa_id = pmd2.numa_id
    fx_port3 = pmd3.add_port(port3_name)
    fx_port3.numa_id = pmd3.numa_id
    fx_port4 = pmd4.add_port(port4_name)
    fx_port4.numa_id = pmd4.numa_id
    fx_port5 = pmd1.add_port(port5_name)
    fx_port5.numa_id = pmd1.numa_id
    fx_port6 = pmd2.add_port(port6_name)
    fx_port6.numa_id = pmd2.numa_id
    fx_port7 = pmd3.add_port(port7_name)
    fx_port7.numa_id = pmd3.numa_id
    fx_port8 = pmd4.add_port(port8_name)
    fx_port8.numa_id = pmd4.numa_id

    # add a dummy rxq into port.
    fx_p1rxq = fx_port1.add_rxq(0)
    fx_p1rxq.pmd = pmd1
    fx_p2rxq = fx_port2.add_rxq(0)
    fx_p2rxq.pmd = pmd2
    fx_p3rxq = fx_port3.add_rxq(0)
    fx_p3rxq.pmd = pmd3
    fx_p4rxq = fx_port4.add_rxq(0)
    fx_p4rxq.pmd = pmd4
    fx_p5rxq = fx_port5.add_rxq(0)
    fx_p5rxq.pmd = pmd1
    fx_p6rxq = fx_port6.add_rxq(0)
    fx_p6rxq.pmd = pmd2
    fx_p7rxq = fx_port7.add_rxq(0)
    fx_p7rxq.pmd = pmd3
    fx_p8rxq = fx_port8.add_rxq(0)
    fx_p8rxq.pmd = pmd4

    # add some cpu consumption for these rxqs.
    # order of rxqs based on cpu consumption: rxqp1,rxqp2,..rxqp8
    for i in range(0, config.ncd_samples_max):
        fx_p1rxq.cpu_cyc[i] = 76
        fx_p1rxq.rx_cyc[i] = 76
        fx_p2rxq.cpu_cyc[i] = 75
        fx_p2rxq.rx_cyc[i] = 75
        fx_p3rxq.cpu_cyc[i] = 74
        fx_p3rxq.rx_cyc[i] = 74
        fx_p4rxq.cpu_cyc[i] = 73
        fx_p4rxq.rx_cyc[i] = 73
        fx_p5rxq.cpu_cyc[i] = 20
        fx_p5rxq.rx_cyc[i] = 20
        fx_p6rxq.cpu_cyc[i] = 15
        fx_p6rxq.rx_cyc[i] = 15
        fx_p7rxq.cpu_cyc[i] = 11
        fx_p7rxq.rx_cyc[i] = 11
        fx_p8rxq.cpu_cyc[i] = 7
        fx_p8rxq.rx_cyc[i] = 7


class TestRebalDryrun_FourPmd(TestCase):
    """
    Test rebalance for one or more rxq handled by four pmds.
    """

    rebalance_dryrun = dataif.rebalance_dryrun_by_cyc
    pmd_map = dict()
    core1_id = 0
    core2_id = 1
    core3_id = 4
    core4_id = 5

    # setup test environment
    def setUp(self):
        util.Memoize.forgot = True

        # turn off limited info shown in assert failure for pmd object.
        self.maxDiff = None

        dataif.Context.nlog = NlogNoop()

        # create one pmd object.
        fx_pmd1 = dataif.Dataif_Pmd(self.core1_id)
        fx_pmd2 = dataif.Dataif_Pmd(self.core2_id)
        fx_pmd3 = dataif.Dataif_Pmd(self.core3_id)
        fx_pmd4 = dataif.Dataif_Pmd(self.core4_id)

        # let it be in numa 0.
        fx_pmd1.numa_id = 0
        fx_pmd2.numa_id = 0
        fx_pmd3.numa_id = 0
        fx_pmd4.numa_id = 0

        # add some cpu consumption for these pmds.
        for i in range(0, config.ncd_samples_max):
            fx_pmd1.idle_cpu_cyc[i] = (4 * (i + 1))
            fx_pmd1.proc_cpu_cyc[i] = (96 * (i + 1))
            fx_pmd1.rx_cyc[i] = (96 * (i + 1))

        fx_pmd1.cyc_idx = config.ncd_samples_max - 1

        for i in range(0, config.ncd_samples_max):
            fx_pmd2.idle_cpu_cyc[i] = (10 * (i + 1))
            fx_pmd2.proc_cpu_cyc[i] = (90 * (i + 1))
            fx_pmd2.rx_cyc[i] = (90 * (i + 1))

        fx_pmd2.cyc_idx = config.ncd_samples_max - 1

        for i in range(0, config.ncd_samples_max):
            fx_pmd3.idle_cpu_cyc[i] = (15 * (i + 1))
            fx_pmd3.proc_cpu_cyc[i] = (85 * (i + 1))
            fx_pmd3.rx_cyc[i] = (85 * (i + 1))

        fx_pmd3.cyc_idx = config.ncd_samples_max - 1

        for i in range(0, config.ncd_samples_max):
            fx_pmd4.idle_cpu_cyc[i] = (20 * (i + 1))
            fx_pmd4.proc_cpu_cyc[i] = (80 * (i + 1))
            fx_pmd4.rx_cyc[i] = (80 * (i + 1))

        fx_pmd4.cyc_idx = config.ncd_samples_max - 1

        self.pmd_map[self.core1_id] = fx_pmd1
        self.pmd_map[self.core2_id] = fx_pmd2
        self.pmd_map[self.core3_id] = fx_pmd3
        self.pmd_map[self.core4_id] = fx_pmd4
        return

    # Test case:
    #   With four threads from same numa, where each pmd thread is handling
    #   two single-queued ports. check whether rebalance is performed.
    #   Scope is to check if rxq from a pmd which was a rebalancing pmd
    #   before, is assigned other pmd successfully.
    #
    #   order of rxqs based on cpu consumption: rxqp1,rxqp2,rxqp3,rxqp4
    #   order of pmds for rebalance dryrun: pmd1,pmd2,pmd3,pmd4
    #
    #   1. rxqp1(pmd1) -NOREB-> rxqp1(pmd1)
    #      rxqp2(pmd2) -NOREB-> rxqp2(pmd2)
    #      rxqp3(pmd3) -NOREB-> rxqp3(pmd3)
    #      rxqp4(pmd4) -NOREB-> rxqp4(pmd4)
    #      rxqp5(pmd1)
    #      rxqp6(pmd2)
    #      rxqp7(pmd3)
    #      rxqp8(pmd4)
    #
    #   2. rxqp1(pmd1) -NOREB-> rxqp1(pmd1)
    #      rxqp2(pmd2) -NOREB-> rxqp2(pmd2)
    #      rxqp3(pmd3) -NOREB-> rxqp3(pmd3)
    #      rxqp4(pmd4) -NOREB-> rxqp4(pmd4)
    #      rxqp5(pmd1) --+--+-> rxqp5(reb_pmd4)
    #      rxqp6(pmd2) --+--+-> rxqp6(reb_pmd3)
    #      rxqp7(pmd3) --+--+-> rxqp7(reb_pmd2)
    #      rxqp8(pmd4) --+--+-> rxqp8(reb_pmd1)
    #
    @mock.patch('netcontrold.lib.util.open')
    def test_eight_1rxq_lnuma(self, mock_open):
        mock_open.side_effect = [
            mock.mock_open(read_data=_FX_4X2CPU_INFO).return_value
        ]

        # set same numa for pmds
        pmd1 = self.pmd_map[self.core1_id]
        pmd2 = self.pmd_map[self.core2_id]
        pmd3 = self.pmd_map[self.core3_id]
        pmd4 = self.pmd_map[self.core4_id]
        pmd1.numa_id = 0
        pmd2.numa_id = 0
        pmd3.numa_id = 0
        pmd4.numa_id = 0

        # create rxq
        fx_4pmd_each_2rxq(self)

        # update pmd load values
        dataif.update_pmd_load(self.pmd_map)

        # copy original pmd objects
        pmd_map = copy.deepcopy(self.pmd_map)

        # test dryrun
        n_reb_rxq = type(self).rebalance_dryrun(self.pmd_map)

        # validate results
        # 1. all four rxqs be rebalanced.
        self.assertEqual(n_reb_rxq, 4, "four rxqs to be rebalanced")
        # 2. each pmd is updated.
        self.assertNotEqual(pmd_map[self.core1_id], pmd1)
        self.assertNotEqual(pmd_map[self.core2_id], pmd2)
        self.assertNotEqual(pmd_map[self.core3_id], pmd3)
        self.assertNotEqual(pmd_map[self.core4_id], pmd4)
        # 3. check rxq map after dryrun.
        port1 = pmd1.find_port_by_name('virtport1')
        port2 = pmd2.find_port_by_name('virtport2')
        port3 = pmd3.find_port_by_name('virtport3')
        port4 = pmd4.find_port_by_name('virtport4')
        port5 = pmd1.find_port_by_name('virtport5')
        port6 = pmd2.find_port_by_name('virtport6')
        port7 = pmd3.find_port_by_name('virtport7')
        port8 = pmd4.find_port_by_name('virtport8')
        self.assertEqual(port1.rxq_rebalanced, {})
        self.assertEqual(port2.rxq_rebalanced, {})
        self.assertEqual(port3.rxq_rebalanced, {})
        self.assertEqual(port4.rxq_rebalanced, {})
        self.assertEqual(port5.rxq_rebalanced[0], pmd4.id)
        self.assertEqual(port6.rxq_rebalanced[0], pmd3.id)
        self.assertEqual(port7.rxq_rebalanced[0], pmd2.id)
        self.assertEqual(port8.rxq_rebalanced[0], pmd1.id)
        # 4. check pmd load
        self.assertEqual(dataif.pmd_load(pmd1), (96.0 - 20.0 + 7.0))
        self.assertEqual(dataif.pmd_load(pmd2), (90.0 - 15.0 + 11.0))
        self.assertEqual(dataif.pmd_load(pmd3), (85.0 - 11.0 + 15.0))
        self.assertEqual(dataif.pmd_load(pmd4), (80.0 - 7.0 + 20.0))

        # del port object from pmd.
        # TODO: create fx_ post deletion routine for clean up
        pmd1.del_port('virtport1')
        pmd2.del_port('virtport2')
        pmd3.del_port('virtport3')
        pmd4.del_port('virtport4')


# Fixture:
#   Create two pmd thread objects per numa where in, one pmd has three
#   single-queued ports, while the other is idle (without any port/rxq).
def fx_2pmd_one_empty_per_numa(testobj):
    # retrieve pmd object.
    pmd1 = testobj.pmd_map[testobj.core1_id]
    pmd3 = testobj.pmd_map[testobj.core3_id]

    # dummy ports required for this test.
    port1_name = 'virtport1'
    port2_name = 'virtport2'
    port3_name = 'virtport3'
    port4_name = 'virtport4'
    port5_name = 'virtport5'
    port6_name = 'virtport6'

    # create port class of name 'virtport'.
    dataif.make_dataif_port(port1_name)
    dataif.make_dataif_port(port2_name)
    dataif.make_dataif_port(port3_name)
    dataif.make_dataif_port(port4_name)
    dataif.make_dataif_port(port5_name)
    dataif.make_dataif_port(port6_name)

    # add port object into pmd.
    fx_port1 = pmd1.add_port(port1_name)
    fx_port1.numa_id = pmd1.numa_id
    fx_port2 = pmd1.add_port(port2_name)
    fx_port2.numa_id = pmd1.numa_id
    fx_port3 = pmd1.add_port(port3_name)
    fx_port3.numa_id = pmd1.numa_id
    fx_port4 = pmd3.add_port(port4_name)
    fx_port4.numa_id = pmd3.numa_id
    fx_port5 = pmd3.add_port(port5_name)
    fx_port5.numa_id = pmd3.numa_id
    fx_port6 = pmd3.add_port(port6_name)
    fx_port6.numa_id = pmd3.numa_id

    # add a dummy rxq into port.
    fx_p1rxq = fx_port1.add_rxq(0)
    fx_p1rxq.pmd = pmd1
    fx_p2rxq = fx_port2.add_rxq(0)
    fx_p2rxq.pmd = pmd1
    fx_p3rxq = fx_port3.add_rxq(0)
    fx_p3rxq.pmd = pmd1
    fx_p4rxq = fx_port4.add_rxq(0)
    fx_p4rxq.pmd = pmd3
    fx_p5rxq = fx_port5.add_rxq(0)
    fx_p5rxq.pmd = pmd3
    fx_p6rxq = fx_port6.add_rxq(0)
    fx_p6rxq.pmd = pmd3

    # add some cpu consumption for these rxqs.
    # order of rxqs based on cpu consumption: rxqp2,rxqp1,rxqp3,
    #                                         rxqp5,rxqp4,rxqp6
    for i in range(0, config.ncd_samples_max):
        fx_p2rxq.cpu_cyc[i] = 66
        fx_p1rxq.cpu_cyc[i] = 20
        fx_p3rxq.cpu_cyc[i] = 10
        fx_p5rxq.cpu_cyc[i] = 66
        fx_p4rxq.cpu_cyc[i] = 20
        fx_p6rxq.cpu_cyc[i] = 10

        fx_p2rxq.rx_cyc[i] = 66
        fx_p1rxq.rx_cyc[i] = 20
        fx_p3rxq.rx_cyc[i] = 10
        fx_p5rxq.rx_cyc[i] = 66
        fx_p4rxq.rx_cyc[i] = 20
        fx_p6rxq.rx_cyc[i] = 10


class TestRebalDryrun_FourPmd_Numa(TestCase):
    """
    Test rebalance for one or more rxq handled by four pmds.
    """

    rebalance_dryrun = dataif.rebalance_dryrun_by_cyc
    pmd_map = dict()
    core1_id = 0
    core2_id = 1
    core3_id = 6
    core4_id = 7

    # setup test environment
    def setUp(self):
        util.Memoize.forgot = True

        # turn off limited info shown in assert failure for pmd object.
        self.maxDiff = None

        dataif.Context.nlog = NlogNoop()

        # create one pmd object.
        fx_pmd1 = dataif.Dataif_Pmd(self.core1_id)
        fx_pmd2 = dataif.Dataif_Pmd(self.core2_id)
        fx_pmd3 = dataif.Dataif_Pmd(self.core3_id)
        fx_pmd4 = dataif.Dataif_Pmd(self.core4_id)

        # let it be in numa 0.
        fx_pmd1.numa_id = 0
        fx_pmd2.numa_id = 0
        fx_pmd3.numa_id = 1
        fx_pmd4.numa_id = 1

        # add some cpu consumption for these pmds.
        for i in range(0, config.ncd_samples_max):
            fx_pmd1.idle_cpu_cyc[i] = (4 * (i + 1))
            fx_pmd1.proc_cpu_cyc[i] = (96 * (i + 1))
            fx_pmd1.rx_cyc[i] = (96 * (i + 1))

        fx_pmd1.cyc_idx = config.ncd_samples_max - 1

        for i in range(0, config.ncd_samples_max):
            fx_pmd2.idle_cpu_cyc[i] = (10 * (i + 1))
            fx_pmd2.proc_cpu_cyc[i] = (90 * (i + 1))
            fx_pmd2.rx_cyc[i] = (90 * (i + 1))

        fx_pmd2.cyc_idx = config.ncd_samples_max - 1

        for i in range(0, config.ncd_samples_max):
            fx_pmd3.idle_cpu_cyc[i] = (15 * (i + 1))
            fx_pmd3.proc_cpu_cyc[i] = (85 * (i + 1))
            fx_pmd3.rx_cyc[i] = (85 * (i + 1))

        fx_pmd3.cyc_idx = config.ncd_samples_max - 1

        for i in range(0, config.ncd_samples_max):
            fx_pmd4.idle_cpu_cyc[i] = (20 * (i + 1))
            fx_pmd4.proc_cpu_cyc[i] = (80 * (i + 1))
            fx_pmd4.rx_cyc[i] = (80 * (i + 1))

        fx_pmd4.cyc_idx = config.ncd_samples_max - 1

        self.pmd_map[self.core1_id] = fx_pmd1
        self.pmd_map[self.core2_id] = fx_pmd2
        self.pmd_map[self.core3_id] = fx_pmd3
        self.pmd_map[self.core4_id] = fx_pmd4
        return

    # Test case:
    #   With two threads per numa, where one pmd thread is handling
    #   two single-queued ports, while the other pmd is empty,
    #   check whether rebalance is performed in each numa.
    #   Scope is to check if only one rxq is moved to empty pmd
    #   within numa affinity.
    #
    #   order of rxqs based on cpu consumption: rxqp2,rxqp1,rxqp5,rxqp4
    #   order of pmds for rebalance dryrun: pmd1N0,pmd3N1,pmd2N0,pmd4N1
    #
    #   1. rxqp2(pmd1N0) -NOREB-> rxqp2(pmd1N0)
    #      rxqp1(pmd1N0)
    #        -  (pmd2N0)
    #
    #   2  rxqp5(pmd3N1) -NOREB-> rxqp5(pmd3N0)
    #      rxqp4(pmd3N1)
    #        -  (pmd4N1)
    #
    #   3. rxqp2(pmd1N0) -NOREB-> rxqp2(pmd1N0)
    #      rxqp1(pmd1N0) --+--+-> rxqp1(reb_pmd2N0)
    #
    #   4. rxqp5(pmd3N1) -NOREB-> rxqp5(pmd3N1)
    #      rxqp4(pmd3N1) --+--+-> rxqp4(reb_pmd4N1)
    #
    @mock.patch('netcontrold.lib.util.open')
    def test_two_1rxq_with_empty_per_numa(self, mock_open):
        mock_open.side_effect = [
            mock.mock_open(read_data=_FX_4X2CPU_INFO).return_value
        ]

        # set numa for pmds
        self.core1_id = 0
        self.core2_id = 1
        self.core3_id = 6
        self.core4_id = 7
        pmd1 = self.pmd_map[self.core1_id]
        pmd2 = self.pmd_map[self.core2_id]
        pmd3 = self.pmd_map[self.core3_id]
        pmd4 = self.pmd_map[self.core4_id]

        # create rxq
        fx_2pmd_one_empty_per_numa(self)

        # delete excess ports in pmds
        pmd1.del_port('virtport3')
        pmd3.del_port('virtport6')

        # update pmd stats for 2 and 4 to be idle.
        for i in range(0, config.ncd_samples_max):
            pmd2.idle_cpu_cyc[i] = (100 * (i + 1))
            pmd2.proc_cpu_cyc[i] = (0 * (i + 1))
            pmd2.rx_cyc[i] = (0 * (i + 1))

        for i in range(0, config.ncd_samples_max):
            pmd4.idle_cpu_cyc[i] = (100 * (i + 1))
            pmd4.proc_cpu_cyc[i] = (0 * (i + 1))
            pmd4.rx_cyc[i] = (0 * (i + 1))

        # update rxq stats for ports in pmd 1 and pmd 3
        port1 = pmd1.find_port_by_name('virtport1')
        port2 = pmd1.find_port_by_name('virtport2')
        port4 = pmd3.find_port_by_name('virtport4')
        port5 = pmd3.find_port_by_name('virtport5')
        p1rxq = port1.find_rxq_by_id(0)
        p2rxq = port2.find_rxq_by_id(0)
        p4rxq = port4.find_rxq_by_id(0)
        p5rxq = port5.find_rxq_by_id(0)
        for i in range(0, config.ncd_samples_max):
            p2rxq.cpu_cyc[i] = 66
            p2rxq.rx_cyc[i] = 66
            p1rxq.cpu_cyc[i] = 30
            p1rxq.rx_cyc[i] = 30
            p5rxq.cpu_cyc[i] = 66
            p5rxq.rx_cyc[i] = 66
            p4rxq.cpu_cyc[i] = 30
            p4rxq.rx_cyc[i] = 30

        # update pmd load values
        dataif.update_pmd_load(self.pmd_map)

        # copy original pmd objects
        pmd_map = copy.deepcopy(self.pmd_map)

        # test dryrun
        n_reb_rxq = type(self).rebalance_dryrun(self.pmd_map)

        # validate results
        # 1. all two rxqs be rebalanced.
        self.assertEqual(n_reb_rxq, 2, "two rxqs to be rebalanced")
        # 2. each pmd is updated.
        self.assertNotEqual(pmd_map[self.core1_id], pmd1)
        self.assertNotEqual(pmd_map[self.core2_id], pmd2)
        self.assertNotEqual(pmd_map[self.core3_id], pmd3)
        self.assertNotEqual(pmd_map[self.core4_id], pmd4)
        # 3. check rxq map after dryrun.
        port2reb = pmd2.find_port_by_name('virtport1')
        port4reb = pmd4.find_port_by_name('virtport4')
        # 3.a rxqp2 remains in pmd1
        self.assertEqual(port2.rxq_rebalanced, {})
        self.assertEqual(port2.find_rxq_by_id(0).pmd.id, pmd1.id)
        # 3.b rxqp3 remains in pmd3
        self.assertEqual(port5.rxq_rebalanced, {})
        self.assertEqual(port5.find_rxq_by_id(0).pmd.id, pmd3.id)
        # 3.c rxqp1 moves from pmd1 to pmd2
        self.assertEqual(port1.rxq_rebalanced[0], pmd2.id)
        self.assertIsNone(port1.find_rxq_by_id(0))
        # 3.c.0 and dry-run did not break original pinning.
        rxqp2reb = port2reb.find_rxq_by_id(0)
        self.assertEqual(rxqp2reb.pmd.id, pmd1.id)
        # 3.d rxqp4 moves from pmd3 to pmd4
        self.assertEqual(port4.rxq_rebalanced[0], pmd4.id)
        self.assertIsNone(port4.find_rxq_by_id(0))
        # 3.d.0 and dry-run did not break original pinning.
        rxqp4reb = port4reb.find_rxq_by_id(0)
        self.assertEqual(rxqp4reb.pmd.id, pmd3.id)
        # 4. check pmd load
        self.assertEqual(dataif.pmd_load(pmd1), (96.0 - 30.0))
        self.assertEqual(dataif.pmd_load(pmd2), 30.0)
        self.assertEqual(dataif.pmd_load(pmd3), (85.0 - 30.0))
        self.assertEqual(dataif.pmd_load(pmd4), 30.0)

        # del port object from pmd.
        # TODO: create fx_ post deletion routine for clean up
        pmd1.del_port('virtport1')
        pmd1.del_port('virtport2')
        pmd3.del_port('virtport5')
        pmd3.del_port('virtport4')

    # Test case:
    #   With two threads per numa, where one pmd thread is handling
    #   three single-queued ports in first numa, with the other pmd being
    #   idle, at the same time pmds in other numa are entirely idle.
    #   Check whether rebalance is performed in only first numa.
    #   Scope is to check if only one rxq is moved to empty pmd
    #   within numa affinity.
    #
    #   order of rxqs based on cpu consumption: rxqp2,rxqp1,rxqp3
    #   order of pmds for rebalance dryrun: pmd1N0,pmd3N1,pmd2N0,pmd4N1
    #
    #   1. rxqp2(pmd1N0) -NOREB-> rxqp2(pmd1N0)
    #      rxqp1(pmd1N0)
    #      rxqp3(pmd1N0)
    #        -  (pmd2N0)
    #
    #        -  (pmd3N1)
    #        -  (pmd4N1)
    #
    #   2. rxqp2(pmd1N0) -NOREB-> rxqp2(pmd1N0)
    #      rxqp1(pmd1N0) --+--+-> rxqp1(reb_pmd2N0)
    #      rxqp3(pmd1N0)
    #
    #        -  (pmd3N1)
    #        -  (pmd4N1)
    #
    #   3. rxqp2(pmd1N0) -NOREB-> rxqp2(pmd1N0)
    #      rxqp1(pmd1N0) --+--+-> rxqp1(reb_pmd2N0)
    #      rxqp3(pmd1N0) --+--+-> rxqp3(reb_pmd2N0)
    #
    #        -  (pmd3N1)
    #        -  (pmd4N1)
    #
    @mock.patch('netcontrold.lib.util.open')
    def test_two_1rxq_with_empty_one_numa(self, mock_open):
        mock_open.side_effect = [
            mock.mock_open(read_data=_FX_4X2CPU_INFO).return_value
        ]

        # set numa for pmds
        self.core1_id = 0
        self.core2_id = 1
        self.core3_id = 6
        self.core4_id = 7
        pmd1 = self.pmd_map[self.core1_id]
        pmd2 = self.pmd_map[self.core2_id]
        pmd3 = self.pmd_map[self.core3_id]
        pmd4 = self.pmd_map[self.core4_id]

        # create rxq
        fx_2pmd_one_empty_per_numa(self)

        # empty pmd threads in second numa
        pmd3.del_port('virtport4')
        pmd3.del_port('virtport5')
        pmd3.del_port('virtport6')

        # update pmd stats for 2, 3 and 4 to be idle.
        for i in range(0, config.ncd_samples_max):
            pmd2.idle_cpu_cyc[i] = (100 * (i + 1))
            pmd2.proc_cpu_cyc[i] = (0 * (i + 1))
            pmd2.rx_cyc[i] = (0 * (i + 1))

            pmd3.idle_cpu_cyc[i] = (100 * (i + 1))
            pmd3.proc_cpu_cyc[i] = (0 * (i + 1))
            pmd3.rx_cyc[i] = (0 * (i + 1))

            pmd4.idle_cpu_cyc[i] = (100 * (i + 1))
            pmd4.proc_cpu_cyc[i] = (0 * (i + 1))
            pmd4.rx_cyc[i] = (0 * (i + 1))

        # update pmd load values
        dataif.update_pmd_load(self.pmd_map)

        # copy original pmd objects
        pmd_map = copy.deepcopy(self.pmd_map)

        # test dryrun
        n_reb_rxq = type(self).rebalance_dryrun(self.pmd_map)

        # validate results
        # 1. two rxqs be rebalanced in numa 1.
        self.assertEqual(n_reb_rxq, 2, "two rxqs to be rebalanced")
        # 2. each pmd is updated, except numa 2.
        self.assertNotEqual(pmd_map[self.core1_id], pmd1)
        self.assertNotEqual(pmd_map[self.core2_id], pmd2)
        self.assertEqual(pmd_map[self.core3_id], pmd3)
        self.assertEqual(pmd_map[self.core4_id], pmd4)
        # 3. check rxq map after dryrun.
        port1 = pmd1.find_port_by_name('virtport1')
        port2 = pmd1.find_port_by_name('virtport2')
        port3 = pmd1.find_port_by_name('virtport3')
        port1reb = pmd2.find_port_by_name('virtport1')
        port3reb = pmd2.find_port_by_name('virtport3')
        # 3.a rxqp2 remains in pmd1
        self.assertEqual(port2.rxq_rebalanced, {})
        self.assertEqual(port2.find_rxq_by_id(0).pmd.id, pmd1.id)
        # 3.b rxqp1 moves from pmd1 to pmd2
        self.assertEqual(port1.rxq_rebalanced[0], pmd2.id)
        self.assertIsNone(port1.find_rxq_by_id(0))
        # 3.b.0 and dry-run did not break original pinning.
        rxqp1reb = port1reb.find_rxq_by_id(0)
        self.assertEqual(rxqp1reb.pmd.id, pmd1.id)
        # 3.c rxqp3 moves from pmd1 to pmd2
        self.assertEqual(port3.rxq_rebalanced[0], pmd2.id)
        self.assertIsNone(port3.find_rxq_by_id(0))
        # 3.c.0 and dry-run did not break original pinning.
        rxqp3reb = port3reb.find_rxq_by_id(0)
        self.assertEqual(rxqp3reb.pmd.id, pmd1.id)
        # 3.d no port moved into numa 1
        self.assertEqual(pmd3.count_rxq(), 0)
        self.assertEqual(pmd4.count_rxq(), 0)
        # 4. check pmd load
        self.assertEqual(dataif.pmd_load(pmd1), (96.0 - 20.0 - 10.0))
        self.assertEqual(dataif.pmd_load(pmd2), 30.0)
        self.assertEqual(dataif.pmd_load(pmd3), 0)
        self.assertEqual(dataif.pmd_load(pmd4), 0)

        # del port object from pmd.
        # TODO: create fx_ post deletion routine for clean up
        pmd1.del_port('virtport1')
        pmd1.del_port('virtport2')


class TestRebalDryrunIQ_OnePmd(TestRebalDryrun_OnePmd):
    """
    Test rebalance for one or more rxq handled by one pmd.
    """

    rebalance_dryrun = dataif.rebalance_dryrun_by_iq
    pmd_map = dict()
    core_id = 0


class TestRebalDryrunIQ_TwoPmd(TestRebalDryrun_TwoPmd):
    """
    Test rebalance for one or more rxq handled by two pmds.
    """

    rebalance_dryrun = dataif.rebalance_dryrun_by_iq
    pmd_map = dict()
    core1_id = 0
    core2_id = 1

    @pytest.mark.skip(reason="not applicable")
    def test_four_1rxq_lnuma(self, mock_open):
        ...

    @pytest.mark.skip(reason="not applicable")
    def test_four_1rxq_skip_lnuma(self, mock_open):
        ...

    @pytest.mark.skip(reason="not applicable")
    def test_two_1p2rxq_lnuma(self, mock_open):
        ...

    @pytest.mark.skip(reason="not applicable")
    def test_two_1p2rxq_lnuma_norb(self, mock_open):
        ...


class TestRebalDryrunIQ_FourPmd(TestRebalDryrun_FourPmd):
    """
    Test rebalance for one or more rxq handled by four pmds.
    """

    rebalance_dryrun = dataif.rebalance_dryrun_by_iq
    pmd_map = dict()
    core1_id = 0
    core2_id = 1
    core3_id = 4
    core4_id = 5

    @pytest.mark.skip(reason="not applicable")
    def test_eight_1rxq_lnuma(self, mock_open):
        ...


class TestRebalDryrunIQ_FourPmd_Numa(TestRebalDryrun_FourPmd_Numa):
    """
    Test rebalance for one or more rxq handled by four pmds.
    """

    rebalance_dryrun = dataif.rebalance_dryrun_by_cyc
    pmd_map = dict()
    core1_id = 0
    core2_id = 1
    core3_id = 6
    core4_id = 7


class TestRebalDryrunVarLoad_TwoPmd(TestCase):
    """
    Test rebalance for one or more rxq handled by two pmds.
    """

    rebalance_dryrun = dataif.rebalance_dryrun_by_cyc
    pmd_map = dict()
    core1_id = 0
    core2_id = 1

    # setup test environment
    def setUp(self):
        util.Memoize.forgot = True

        # turn off limited info shown in assert failure for pmd object.
        self.maxDiff = None

        dataif.Context.nlog = NlogNoop()

        # create one pmd object.
        fx_pmd1 = dataif.Dataif_Pmd(self.core1_id)
        fx_pmd2 = dataif.Dataif_Pmd(self.core2_id)

        # let it be in numa 0.
        fx_pmd1.numa_id = 0
        fx_pmd2.numa_id = 0

        # add some cpu consumption for these pmds.
        # pmd1 to be 95% busy i.e using 5*96 cycles out of 5*100 cycles
        for i in range(0, config.ncd_samples_max):
            fx_pmd1.idle_cpu_cyc[i] = (4 * (i + 1))
            fx_pmd1.proc_cpu_cyc[i] = (96 * (i + 1))
            fx_pmd1.rx_cyc[i] = (96 * (i + 1))
        fx_pmd1.cyc_idx = config.ncd_samples_max - 1

        # pmd2 to be 90% busy i.e using 5*90 cycles out of 5*100 cycles
        for i in range(0, config.ncd_samples_max):
            fx_pmd2.idle_cpu_cyc[i] = (10 * (i + 1))
            fx_pmd2.proc_cpu_cyc[i] = (90 * (i + 1))
            fx_pmd2.rx_cyc[i] = (90 * (i + 1))
        fx_pmd2.cyc_idx = config.ncd_samples_max - 1

        self.pmd_map[self.core1_id] = fx_pmd1
        self.pmd_map[self.core2_id] = fx_pmd2
        return

    # Test case:
    #   With two threads from same numa, where one pmd thread is handling
    #   two single-queued ports, while the other pmd is empty,
    #   check whether rebalance is performed.
    #   Scope is to check if only one rxq is moved to empty pmd.
    #
    #   order of rxqs based on cpu consumption: rxqp2,rxqp1
    #   order of pmds for rebalance dryrun: pmd1,pmd2
    #
    #   1. rxqp2(pmd1) -NOREB-> rxqp2(pmd1)
    #      rxqp1(pmd1)
    #        -  (pmd2)
    #
    #   2. rxqp2(pmd1) -NOREB-> rxqp2(pmd1)
    #      rxqp1(pmd1) --+--+-> rxqp1(reb_pmd2)
    #
    @mock.patch('netcontrold.lib.util.open')
    def test_two_1rxq_with_empty_lnuma(self, mock_open):
        mock_open.side_effect = [
            mock.mock_open(read_data=_FX_CPU_INFO).return_value
        ]

        # set same numa for pmds
        pmd1 = self.pmd_map[self.core1_id]
        pmd2 = self.pmd_map[self.core2_id]
        pmd1.numa_id = 0
        pmd2.numa_id = 0

        # let pmd2 be idle
        for i in range(0, config.ncd_samples_max):
            pmd2.idle_cpu_cyc[i] = (100 * (i + 1))
            pmd2.proc_cpu_cyc[i] = (0 * (i + 1))
            pmd2.rx_cyc[i] = (0 * (i + 1))

        # create rxq
        fx_2pmd_one_empty(self)

        # update rxq stats
        port1 = pmd1.find_port_by_name('virtport1')
        port2 = pmd1.find_port_by_name('virtport2')
        p1rxq = port1.find_rxq_by_id(0)
        p2rxq = port2.find_rxq_by_id(0)

        # fluctuate rxq consumption within 96% load in pmd1
        # p1rxq consumption is 36 % in 5*100 cyc ie 180
        # p2rxq consumption is 60 % in 5*100 cyc ie 300
        p1rxq.rx_cyc = [0, 61, 51, 31, 21, 16]
        p1rxq.cpu_cyc = [0, 61, 51, 31, 21, 16]
        p2rxq.rx_cyc = [0, 35, 45, 65, 75, 80]
        p2rxq.cpu_cyc = [0, 35, 45, 65, 75, 80]

        # update pmd load values
        dataif.update_pmd_load(self.pmd_map)

        # copy original pmd objects
        pmd_map = copy.deepcopy(self.pmd_map)

        # test dryrun
        n_reb_rxq = type(self).rebalance_dryrun(self.pmd_map)

        # validate results
        # 1. all two rxqs be rebalanced.
        self.assertEqual(n_reb_rxq, 1, "one rxq to be rebalanced")
        # 2. each pmd is updated.
        self.assertNotEqual(pmd_map[self.core1_id], pmd1)
        self.assertNotEqual(pmd_map[self.core2_id], pmd2)
        # 3. check rxq map after dryrun.
        # 3.a rxqp2 remains in pmd1
        self.assertEqual(port2.rxq_rebalanced, {})
        self.assertEqual(port2.find_rxq_by_id(0).pmd.id, pmd1.id)
        # 3.a rxqp1 moves into pmd2
        self.assertEqual(port1.rxq_rebalanced[0], pmd2.id)
        # 4. check pmd load
        self.assertEqual(dataif.pmd_load(pmd1), 60.0)
        self.assertEqual(dataif.pmd_load(pmd2), 36.0)

        # del port object from pmd.
        # TODO: create fx_ post deletion routine for clean up
        pmd1.del_port('virtport1')
        pmd1.del_port('virtport2')
